#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides Qt table widgets which display logging messages from the
python :mod:`logging` module"""

from __future__ import absolute_import

from operator import attrgetter
from builtins import range

import logging
import logging.handlers
import datetime
import threading
import socket
import click

import taurus
from taurus.core.util.log import Logger
from taurus.core.util.remotelogmonitor import LogRecordStreamHandler, \
    LogRecordSocketReceiver
from taurus.core.util.decorator.memoize import memoized

from taurus.external.qt import Qt
from taurus.qt.qtgui.model import FilterToolBar
from taurus.qt.qtgui.util import ActionFactory
import taurus.cli.common

from .qtable import QBaseTableWidget


__all__ = ["QLoggingTableModel", "QLoggingTable", "QLoggingWidget",
           "QRemoteLoggingTableModel"]

__docformat__ = 'restructuredtext'

LEVEL, TIME, MSG, NAME, ORIGIN = list(range(5))
HORIZ_HEADER = 'Level', 'Time', 'Message', 'By', 'Origin'

__LEVEL_BRUSH = {
    taurus.Trace: (Qt.Qt.lightGray, Qt.Qt.black),
    taurus.Debug: (Qt.Qt.green, Qt.Qt.black),
    taurus.Info: (Qt.Qt.blue, Qt.Qt.white),
    taurus.Warning: (Qt.QColor(255, 165, 0), Qt.Qt.black),
    taurus.Error: (Qt.Qt.red, Qt.Qt.black),
    taurus.Critical: (Qt.QColor(160, 32, 240), Qt.Qt.white),
}


def getBrushForLevel(level):
    elevel = taurus.Trace
    if level <= taurus.Trace:
        elevel = taurus.Trace
    elif level <= taurus.Debug:
        elevel = taurus.Debug
    elif level <= taurus.Info:
        elevel = taurus.Info
    elif level <= taurus.Warning:
        elevel = taurus.Warning
    elif level <= taurus.Error:
        elevel = taurus.Error
    elif level <= taurus.Critical:
        elevel = taurus.Critical
    f, g = list(map(Qt.QBrush, __LEVEL_BRUSH[elevel]))
    return f, g


gethostname = memoized(socket.gethostname)


def _get_record_origin(rec):
    host = getattr(rec, 'hostName', "?" + gethostname() + "?")
    procName = getattr(rec, 'processName', "?process?")
    procID = getattr(rec, 'process', "?PID?")
    threadName = getattr(rec, 'threadName', "?thread?")
    threadID = getattr(rec, 'thread', "?threadID?")
    return host, procName, procID, threadName, threadID


def _get_record_trace(rec):
    pathname = getattr(rec, 'pathname', '')
    filename = getattr(rec, 'filename', '')
    modulename = getattr(rec, 'module', '')
    funcname = getattr(rec, 'funcName', '')
    lineno = getattr(rec, 'lineno', '')
    return pathname, filename, modulename, funcname, lineno


def _get_record_origin_str(rec):
    return "{0}.{1}.{3}".format(*_get_record_origin(rec))


def _get_record_origin_tooltip(rec):

    host, procName, procID, threadName, threadID = _get_record_origin(rec)
    pathname, filename, modulename, funcname, lineno = _get_record_trace(rec)
    timestamp = str(datetime.datetime.fromtimestamp(rec.created))
    bgcolor, fgcolor = list(map(Qt.QBrush.color, getBrushForLevel(rec.levelno)))
    bgcolor = "#%02x%02x%02x" % (
        bgcolor.red(), bgcolor.green(), bgcolor.blue())
    fgcolor = "#%02x%02x%02x" % (
        fgcolor.red(), fgcolor.green(), fgcolor.blue())

    return """<html><font face="monospace" size="1">
<table border="0" cellpadding="0" cellspacing="0">
<tr><td>Level:</td><td><font color="{level_bgcolor}">{level}</font></td></tr>
<tr><td>Time:</td><td>{timestamp}</td></tr>
<tr><td>Message:</td><td>{message}</td></tr>
<tr><td>By:</td><td>{name}</td></tr>
<tr><td>Host:</td><td>{host}</td></tr>
<tr><td>Process:</td><td>{procname}({procID})</td></tr>
<tr><td>Thread:</td><td>{threadname}({threadID})</td></tr>
<tr><td>From:</td><td>File pathname({filename}), line {lineno}, in {funcname}</td></tr>
</table></font></html>
""".format(level=rec.levelname, level_fgcolor=fgcolor, level_bgcolor=bgcolor,
           timestamp=timestamp, message=rec.getMessage(),
           name=rec.name, host=host, procname=procName, procID=procID,
           threadname=threadName, threadID=threadID,
           pathname=pathname, filename=filename, funcname=funcname,
           lineno=lineno)


class QLoggingTableModel(Qt.QAbstractTableModel, logging.Handler):

    DftFont = Qt.QFont("Mono", 8)
    DftColSize = Qt.QSize(80, 20), Qt.QSize(200, 20), \
        Qt.QSize(300, 20), Qt.QSize(180, 20), Qt.QSize(240, 20),

    def __init__(self, parent=None, capacity=500000, freq=0.25):
        super(Qt.QAbstractTableModel, self).__init__()
        logging.Handler.__init__(self)
        self._capacity = capacity
        self._records = []
        self._accumulated_records = []
        Logger.addRootLogHandler(self)
        self.startTimer(freq * 1000)

    # ---------------------------------
    # Qt.QAbstractTableModel overwrite
    # ---------------------------------

    def sort(self, column, order=Qt.Qt.AscendingOrder):
        column2key_map = {LEVEL: attrgetter('levelno'),
                          TIME: attrgetter('created'),
                          MSG: attrgetter('msg'),
                          NAME: attrgetter('name'),
                          ORIGIN: attrgetter('process', 'thread', 'name'),
        }
        self._records = sorted(self._records, key=column2key_map[column],
                               reverse=order == Qt.Qt.DescendingOrder)

    def rowCount(self, index=Qt.QModelIndex()):
        return len(self._records)

    def columnCount(self, index=Qt.QModelIndex()):
        return len(HORIZ_HEADER)

    def getRecord(self, index):
        return self._records[index.row()]

    def data(self, index, role=Qt.Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < len(self._records)):
            return None
        record = self.getRecord(index)
        column = index.column()
        if role == Qt.Qt.DisplayRole:
            if column == LEVEL:
                return record.levelname
            elif column == TIME:
                dt = datetime.datetime.fromtimestamp(record.created)
                return str(dt)
                # return dt.strftime("%Y-%m-%d %H:%m:%S.%f")
            elif column == MSG:
                return record.getMessage()
            elif column == NAME:
                return record.name
            elif column == ORIGIN:
                return _get_record_origin_str(record)
        elif role == Qt.Qt.TextAlignmentRole:
            if column in (LEVEL, MSG):
                return Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter
            return Qt.Qt.AlignRight | Qt.Qt.AlignVCenter
        elif role == Qt.Qt.BackgroundRole:
            if column == LEVEL:
                return getBrushForLevel(record.levelno)[0]
        elif role == Qt.Qt.ForegroundRole:
            if column == LEVEL:
                return getBrushForLevel(record.levelno)[1]
        elif role == Qt.Qt.ToolTipRole:
            return _get_record_origin_tooltip(record)
        elif role == Qt.Qt.SizeHintRole:
            return self._getSizeHint(column)
        # elif role == Qt.Qt.StatusTipRole:
        # elif role == Qt.Qt.CheckStateRole:
        elif role == Qt.Qt.FontRole:
            return self.DftFont
        return None

    def _getSizeHint(self, column):
        return QLoggingTableModel.DftColSize[column]

    def headerData(self, section, orientation, role=Qt.Qt.DisplayRole):
        if role == Qt.Qt.TextAlignmentRole:
            if orientation == Qt.Qt.Horizontal:
                return int(Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter)
            return int(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter)
        elif role == Qt.Qt.SizeHintRole:
            if orientation == Qt.Qt.Vertical:
                return Qt.QSize(50, 20)
            else:
                return self._getSizeHint(section)
        elif role == Qt.Qt.FontRole:
            return Qt.QFont("Mono", 8)
        elif role == Qt.Qt.ToolTipRole:
            if section == LEVEL:
                return "log level"
            elif section == TIME:
                return "log time stamp"
            elif section == MSG:
                return "log message"
            elif section == NAME:
                return "object who recorded the log"
            elif section == ORIGIN:
                return ("the host, process and thread where the"
                        + " log was executed from")
        if role != Qt.Qt.DisplayRole:
            return None
        if orientation == Qt.Qt.Horizontal:
            return HORIZ_HEADER[section]
        return int(section + 1)

    def insertRows(self, position, rows=1, index=Qt.QModelIndex()):
        self.beginInsertRows(Qt.QModelIndex(), position, position + rows - 1)
        self.endInsertRows()

    def removeRows(self, position, rows=1, index=Qt.QModelIndex()):
        self.beginRemoveRows(Qt.QModelIndex(), position, position + rows - 1)
        self.endRemoveRows()

    # def setData(self, index, value, role=Qt.Qt.DisplayRole):
    #    pass

    # def flags(self, index)
    #    pass

    # def insertColumns(self):
    #    pass

    # def removeColumns(self):
    #    pass

    # --------------------------
    # logging.Handler overwrite
    # --------------------------

    def timerEvent(self, evt):
        self.updatePendingRecords()

    def updatePendingRecords(self):
        if not self._accumulated_records:
            return
        row_nb = self.rowCount()
        records = self._accumulated_records
        self._accumulated_records = []
        self._records.extend(records)
        self.insertRows(row_nb, len(records))
        if len(self._records) > self._capacity:
            start = len(self._records) - self._capacity
            self._records = self._records[start:]
            self.removeRows(0, start)

    def emit(self, record):
        self._accumulated_records.append(record)

    def flush(self):
        pass

    def close(self):
        self.flush()
        del self._records[:]
        logging.Handler.close(self)


class _LogRecordStreamHandler(LogRecordStreamHandler):

    def handleLogRecord(self, record):
        self.server.data.get('model').emit(record)


class QRemoteLoggingTableModel(QLoggingTableModel):
    """A remote Qt table that displays the taurus logging messages"""

    def connect_logging(self, host='localhost',
                        port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                        handler=_LogRecordStreamHandler):
        self.log_receiver = LogRecordSocketReceiver(host=host, port=port,
                                                    handler=handler, model=self)
        self.log_thread = threading.Thread(
            target=self.log_receiver.serve_until_stopped)
        self.log_thread.daemon = False
        self.log_thread.start()

    def disconnect_logging(self):
        if not hasattr(self, 'log_receiver') or self.log_receiver is None:
            return
        self.log_receiver.stop()
        self.log_thread.join()
        del self.log_receiver


class QLoggingTable(Qt.QTableView):
    """A Qt table that displays the taurus logging messages"""

    scrollLock = False

    def rowsInserted(self, index, start, end):
        """Overwrite of slot rows inserted to do proper resize and scroll to
        bottom if desired"""
        Qt.QTableView.rowsInserted(self, index, start, end)
        for i in range(start, end + 1):
            self.resizeRowToContents(i)
        if start == 0:
            self.resizeColumnsToContents()
        if not self.scrollLock:
            self.scrollToBottom()

    def setScrollLock(self, scrollLock):
        """Sets the state for scrollLock"""
        self.scrollLock = scrollLock

    def getScrollLock(self):
        """Returns wheater or not the scrollLock is active"""
        return self.scrollLock

    def resetScrollLock(self):
        self.setScrollLock(QLoggingTable.ScrollLock)


class LoggingToolBar(FilterToolBar):

    scrollLockToggled = Qt.pyqtSignal(bool)

    def __init__(self, view=None, parent=None, designMode=False):
        FilterToolBar.__init__(self, view=view, parent=parent,
                               designMode=designMode)
        self.getFilterLineEdit().setToolTip("Quick filter by log name")

        self._logLevelComboBox = logLevelComboBox = Qt.QComboBox()
        levels = "Trace", "Debug", "Info", "Warning", "Error", "Critical"
        for level in levels:
            logLevelComboBox.addItem(level, getattr(taurus, level))
        logLevelComboBox.setCurrentIndex(0)
        logLevelComboBox.currentIndexChanged.connect(self.onLogLevelChanged)
        logLevelComboBox.setToolTip("Filter by log level")

        self._filterLevelAction = self.addWidget(logLevelComboBox)
        self.addSeparator()

        af = ActionFactory()
        self._scrollLockAction = af.createAction(self, "Refresh",
                                                 icon=Qt.QIcon.fromTheme(
                                                     "system-lock-screen"),
                                                 tip="Scroll lock",
                                                 toggled=self.onToggleScrollLock)

        self.addAction(self._scrollLockAction)

    def onToggleScrollLock(self, yesno):
        self.scrollLockToggled.emit(yesno)

    def onLogLevelChanged(self, index):
        self.onFilterChanged()

    def getLogLevelComboBox(self):
        return self._logLevelComboBox

    def getLogLevel(self):
        combo = self.getLogLevelComboBox()
        return combo.itemData(combo.currentIndex())

    def setLogLevel(self, level):
        combo = self.getLogLevelComboBox()
        for i in range(combo.count()):
            l = combo.itemData(i)
            if l == level:
                combo.setCurrentIndex(i)


class QLoggingFilterProxyModel(Qt.QSortFilterProxyModel):
    """A filter by log record object name"""

    def __init__(self, parent=None):
        Qt.QSortFilterProxyModel.__init__(self, parent)
        self._logLevel = taurus.Trace

        # filter configuration
        self.setFilterCaseSensitivity(Qt.Qt.CaseInsensitive)
        self.setFilterKeyColumn(0)
        self.setFilterRole(Qt.Qt.DisplayRole)

        # sort configuration
        # self.setSortCaseSensitivity(Qt.Qt.CaseInsensitive)
        # self.setSortRole(Qt.Qt.DisplayRole)

        # general configuration

    def setFilterLogLevel(self, level):
        self._logLevel = level

    def __getattr__(self, name):
        return getattr(self.sourceModel(), name)

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, NAME, sourceParent)
        record = self.getRecord(idx)
        if record.levelno < self._logLevel:
            return False
        name = str(sourceModel.data(idx))
        regexp = self.filterRegExp()
        if regexp.indexIn(name) != -1:
            return True
        return False


_W = "Warning: Switching log perspective will erase previous log messages " \
     "from current perspective!"


class QLoggingWidget(QBaseTableWidget):

    KnownPerspectives = {
        'Standard': {
            "label": "Local",
            "icon": "computer",
            "tooltip": "Local logging.\n" + _W,
            "model": [QLoggingFilterProxyModel, QLoggingTableModel, ],
        },
        'Remote': {
            "label": "Remote",
            "icon": "network-server",
            "tooltip": "Monitor remote logs.\n" + _W,
            "model": [QLoggingFilterProxyModel, QRemoteLoggingTableModel, ],
        },
    }

    DftPerspective = 'Standard'

    def __init__(self, parent=None, designMode=False,
                 with_filter_widget=LoggingToolBar,
                 with_selection_widget=True, with_refresh_widget=True,
                 perspective=None, proxy=None):
        QBaseTableWidget.__init__(self, parent=parent, designMode=designMode,
                                  with_filter_widget=with_filter_widget,
                                  with_selection_widget=False, with_refresh_widget=False,
                                  perspective=perspective, proxy=proxy)

    def createViewWidget(self, klass=None):
        if klass is None:
            klass = QLoggingTable
        view = QBaseTableWidget.createViewWidget(self, klass=klass)
        hh = view.horizontalHeader()
        if hh.length() > 0:
            try:
                hh.setSectionResizeMode(MSG, Qt.QHeaderView.Stretch)
            except AttributeError:  # PyQt4
                hh.setResizeMode(MSG, Qt.QHeaderView.Stretch)
        view.setShowGrid(False)
        view.sortByColumn(TIME, Qt.Qt.AscendingOrder)
        return view

    def createToolArea(self):
        tb = QBaseTableWidget.createToolArea(self)
        filterBar = self.getFilterBar()
        filterBar.scrollLockToggled.connect(self.onScrollLockToggled)
        return tb

    def onScrollLockToggled(self, yesno):
        self.viewWidget().setScrollLock(yesno)

    def onFilterChanged(self, filter):
        if not self.usesProxyQModel():
            return
        proxy_model = self.getQModel()
        level = self.getFilterBar().getLogLevel()
        proxy_model.setFilterLogLevel(level)
        return QBaseTableWidget.onFilterChanged(self, filter)

    def onSwitchPerspective(self, perspective):
        self.stop_logging()
        if perspective == "Remote":
            if hasattr(self, 'hostName') and hasattr(self, 'port'):
                host, port = self.hostName, self.port
            else:
                isValid = False
                dft = "%s:%d" % (socket.gethostname(),
                                 logging.handlers.DEFAULT_TCP_LOGGING_PORT)
                while not isValid:
                    txt, res = Qt.QInputDialog.getText(self,
                                                       "Please input remote logging host and port",
                                                       "Location (<host>:<port>):", Qt.QLineEdit.Normal, dft)
                    if not res:
                        return
                    try:
                        host, port = str(txt).split(":", 1)
                        port = int(port)
                        isValid = True
                    except:
                        Qt.QMessageBox.information(self, "Invalid name",
                                                   "Please type a valid <host>:<port>")
            ret = QBaseTableWidget.onSwitchPerspective(self, perspective)
            qmodel = self.getQModel()
            qmodel.connect_logging(host=host, port=port)
        else:
            ret = QBaseTableWidget.onSwitchPerspective(self, perspective)
        return ret

    def destroy(self, destroyWindow=True, destroySubWindows=True):
        self.stop_logging()
        return QBaseTableWidget.destroy(self, destroyWindow, destroySubWindows)

    def stop_logging(self):
        model = self.getBaseQModel()
        if hasattr(model, 'disconnect_logging'):
            model.disconnect_logging()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'module': 'taurus.qt.qtgui.table',
            'group': 'Taurus Views',
            'icon': 'designer:table.png',
            'container': False}


def fill_log():
    import time
    import random

    for i in range(10):
        taurus.info("Hello world %04d" % i)

    loggers = ["Object%02d" % (i + 1) for i in range(10)]
    i = 0
    while True:
        time.sleep(random.random())
        logger = logging.getLogger(random.choice(loggers))
        level = random.randint(taurus.Trace, taurus.Critical)
        logger.log(level, "log message %04d" % i)
        i += 1


def main():
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="Logging demo", app_version="1.0",
                          org_domain="Taurus", org_name="Taurus community")

    taurus.setLogLevel(taurus.Trace)
    taurus.disableLogOutput()
    w = QLoggingWidget()

    taurus.trace("trace message")
    taurus.debug("debug message")
    taurus.info("Hello world")
    taurus.warning("Warning message")
    taurus.error("error message")
    taurus.critical("critical message")
    w.setMinimumSize(1200, 600)
    w.show()
    app.exec_()
    w.stop_logging()


@click.command('qlogmon')
@click.option(
    '--port', 'port', type=int,
    default=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
    show_default=True,
    help='Port where log server is running',
)
@click.option(
    '--log-name', 'log_name',
    default=None,
    help='Filter specific log object',
)
@taurus.cli.common.log_level
def qlogmon_cmd(port, log_name, log_level):
    """Show the Taurus Remote Log Monitor"""
    import taurus
    host = socket.gethostname()
    level = getattr(taurus, log_level.capitalize(), taurus.Trace)

    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(cmd_line_parser=None,
                            app_name="Taurus remote logger")
    w = QLoggingWidget(perspective="Remote")
    w.setMinimumSize(1024, 600)

    filterbar = w.getFilterBar()
    filterbar.setLogLevel(level)
    if log_name is not None:
        filterbar.setFilterText(log_name)
    w.getPerspectiveBar().setEnabled(False)
    w.getQModel().connect_logging(host, port)
    w.show()
    app.exec_()
    w.getQModel().disconnect_logging()


if __name__ == '__main__':
    main()
    # qlogmon_cmd



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""A pure Qt widget designed to be displayed when a real widget cannot be
loaded for any reason (example: a dependency library is not installed)"""

__all__ = ["create_fallback", "create_taurus_fallback", "QFallBackWidget",
           "TaurusFallBackWidget"]

__docformat__ = 'restructuredtext'

import sys
import functools
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseWidget


def create_fallback(widget_klass_name):
    return functools.partial(QFallBackWidget, replaces=widget_klass_name,
                             exc_info=sys.exc_info())


def create_taurus_fallback(widget_klass_name):
    return functools.partial(TaurusFallBackWidget, replaces=widget_klass_name,
                             exc_info=sys.exc_info())


class QFallBackWidget(Qt.QWidget):
    """A FallBack widget to be used when a real widget cannot be loaded for any
    reason (example: a dependency library is not installed)"""

    def __init__(self, replaces=None, parent=None, *args, **kwargs):
        Qt.QWidget.__init__(self, parent)
        if replaces is None:
            replaces = self.__class__.__name__
        self.replaces = replaces
        self.exc_info = exc_info = kwargs.get("exc_info")
        layout = Qt.QVBoxLayout(self)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.setSpacing(2)
        self.setLayout(layout)
        self.label = Qt.QLabel(self)
        self.label.setText("'{0}' could not be displayed".format(replaces))
        layout.addWidget(self.label, 0, Qt.Qt.AlignVCenter)
        if exc_info is not None and exc_info[0] is not None:
            self.details_button = Qt.QPushButton("Details...", self)
            layout.addWidget(self.details_button, 0, Qt.Qt.AlignTop)
            self.details_button.clicked.connect(self.onShowDetails)
        layout.addStretch(1)

    def onShowDetails(self):
        import taurus.qt.qtgui.dialog
        msgbox = taurus.qt.qtgui.dialog.TaurusMessageBox(*self.exc_info,
                                                         parent=self)
        msgbox.setWindowTitle("{0} Error".format(self.replaces))
        msgbox.setText(self.label.text())
        msgbox.exec_()


class TaurusFallBackWidget(QFallBackWidget, TaurusBaseWidget):

    def __init__(self, replaces=None, parent=None, *args, **kwargs):
        self.call__init__(QFallBackWidget, replaces=replaces,
                          parent=parent, *args, **kwargs)
        designMode = kwargs.get("designMode", False)
        self.call__init__(TaurusBaseWidget, replaces, designMode=designMode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a taurus QPushButton based widgets"""
from __future__ import print_function

from builtins import map
from builtins import str
from future.utils import string_types

from taurus.external.qt import Qt, compat
from taurus.core.taurusbasetypes import LockStatus, TaurusLockInfo
from taurus.core.taurusdevice import TaurusDevice
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.core.util import eventfilters
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox


__all__ = ["TaurusLauncherButton", "TaurusCommandButton", "TaurusLockButton"]

__docformat__ = 'restructuredtext'


class _ButtonDialog(Qt.QDialog):
    _widget = None
    deleteWidgetOnClose = False

    def __init__(self, parent=None):
        Qt.QDialog.__init__(self, parent, Qt.Qt.WindowTitleHint)
        l = Qt.QVBoxLayout()
        self.setLayout(l)
        self.previousWidgetConfig = None

    def setWidget(self, widget):
        oldWidget = self.widget()
        if oldWidget is not None:
            try:
                self._widget.setModel(None)
            except:
                pass
            oldWidget.hide()
            oldWidget.setParent(None)
            oldWidget.deleteLater()
        if widget is not None:
            self.layout().addWidget(widget)
        self._widget = widget

    def widget(self):
        return self._widget

    def closeEvent(self, event):
        if self.deleteWidgetOnClose:
            try:
                self.previousWidgetConfig = self.widget().createConfig()
            except:
                self.previousWidgetConfig = None
            self.setWidget(None)
        Qt.QDialog.closeEvent(self, event)


class TaurusLauncherButton(Qt.QPushButton, TaurusBaseWidget):
    '''This class provides a button that launches a modeless dialog containing
    a specified Taurus widget which gets the same model as the button.
    The button does not use the model directly. Instead it passes it to the
    associated widget.

    Code examples::

        # a button that launches a TaurusAttrForm when clicked
        button =  TaurusLauncherButton(widget = TaurusAttrForm())
        button.setModel('a/b/c') #a device name, which will be set at the TaurusAttrForm when clicking

        # a button that launches a taurusLabel (whose model is an attribute: 'a/b/c/attrname')
        button =  TaurusLauncherButton(widget = TaurusLabel())
        button.setModel('a/b/c/attrname') # attr name, which will be set at the TaurusLabel when clicking

    '''

    _widgetClassName = ''
    _args = []
    _kwargs = {}
    _deleteWidgetOnClose = True
    _icon = None
    _text = None

    def __init__(self, parent=None, designMode=False, widget=None, icon=None, text=None):
        '''Constructor

        :param parent: (Qt.QWidget or None) parent of this widget
        :param designMode: (bool) flag for Qt designer
        :param widget: (Qt.QWidget) a QWidget that will be shown when clicking
                       in the button
        :param icon: (Qt.QIcon) icon for the button
        :param text: (str) the button text (if None passed, the widget's class name
                     is used) '''

        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self._dialog = _ButtonDialog(self)
        if icon is None and self._icon is not None:
            icon = Qt.QIcon(self._icon)
        if icon is not None:
            self.setIcon(Qt.QIcon(icon))
        if text is not None:
            self._text = text
        if isinstance(widget, Qt.QWidget):
            # we cannot be sure on recreating the same widget again
            self._deleteWidgetOnClose = False
            self.setWidget(widget)
        elif widget is not None:
            self._widgetClassName = widget
        self.clicked.connect(self.onClicked)
        self.setDefault(False)
        self.setAutoDefault(False)
        # no need to listen to change events!
        self.insertEventFilter(eventfilters.IGNORE_CHANGE_AND_PERIODIC)
        self._updateText()

    def getModelClass(self):
        '''see :meth:`TaurusBaseComponent.getModelClass`. Note that in the case of
        :class:`TaurusLauncherButton`, the class is completely dependent on the
        widget's class'''
        try:
            return self.widget().getModelClass()
        except:
            # return None  #@TODO: Uncommenting this avoids the exception when
            # TaurusBaseWidget.getModelClass chokes with relative classes. But
            # the thing should be solved at TaurusBaseWidget.getModelClass
            # level
            return TaurusBaseWidget.getModelClass(self)

    def setText(self, text):
        '''Sets the text of the button. see :meth:`Qt.QPushButton.setText`'''
        self._text = text
        Qt.QPushButton.setText(self, text)

    def getWidgetClassName(self):
        return self._widgetClassName

    def setWidgetClassName(self, className, args=None, kwargs=None):
        self._widgetClassName = str(className)
        if args is not None:
            self._args = args
        if kwargs is not None:
            self._kwargs = kwargs
        self._updateText()

    def resetWidgetClassName(self, className, args=None, kwargs=None):
        self.setWidgetClassName(self.__class__._widgetClassName)

    def createWidget(self):
        from taurus.qt.qtgui.util import TaurusWidgetFactory
        klass = TaurusWidgetFactory().getWidgetClass(self._widgetClassName)
        widget = klass(*self._args, **self._kwargs)
        self.setWidget(widget)
        if self._dialog.previousWidgetConfig is not None:
            try:
                widget.applyConfig(self._dialog.previousWidgetConfig)
            except Exception as e:
                self.warning(
                    'Cannot apply previous configuration to widget. Reason: %s', repr(e))

    def widget(self):
        return self._dialog.widget()

    def setWidget(self, widget):
        '''sets the widget that will be shown when clicking the button

        :param widget: (Qt.QWidget)
        '''
        self._dialog.setWidget(widget)
        self._updateText()

    def displayValue(self, v):
        '''see :meth:`TaurusBaseComponent.displayValue`'''
        if self._text is not None:
            return self._text  # make sure the text is not changed once set
        TaurusBaseWidget.displayValue(self, v)

    def getDisplayValue(self):
        '''see :meth:`TaurusBaseComponent.getDisplayValue`'''
        if self._text is not None:
            return self._text
        if self.widget() is not None:
            return self.widget().__class__.__name__
        return self._widgetClassName or '---'

    def _updateText(self):
        Qt.QPushButton.setText(self, self.getDisplayValue())

    def onClicked(self):
        '''
        Slot called when the button is clicked.
        Note that the dialog will only be created once. Subsequent clicks on
        the button will only raise the existing dialog'''
        if self.widget() is None:
            self.createWidget()
        self.widget().setModel(self.getModelName())
        self._dialog.deleteWidgetOnClose = self._deleteWidgetOnClose
        # dialog.resize(400,300)

        # It's always necessary to set the title since the model could have
        # changed
        self._dialog.setWindowTitle(str(self.getModelName()))

        self._dialog.show()
        self._dialog.raise_()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'group': 'Taurus Buttons',
            'icon': 'designer:pushbutton.png',
            'module': 'taurus.qt.qtgui.button',
            'container': False}

    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    #                      Qt Properties                        #
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    Model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel, TaurusBaseWidget.resetModel)
    UseParentModel = Qt.pyqtProperty("bool", TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel, TaurusBaseWidget.resetUseParentModel)
    widgetClassName = Qt.pyqtProperty(
        "QString", getWidgetClassName, setWidgetClassName, resetWidgetClassName)


class TaurusCommandButton(Qt.QPushButton, TaurusBaseWidget):
    '''This class provides a button that executes a tango command on its device.

    Code examples::

        # a button that executes the "status" command for the 'a/b/c' device server
        button =  TaurusCommandButton(command = 'Status', icon='logos:taurus.png')
        button.setModel('a/b/c')

        #  a button that executes the "exec" command for the 'a/b/c' device server with one parameter
        button =  TaurusCommandButton(command = 'Status', parameters=['2+2'],icon='logos:taurus.png')
        button.setModel('a/b/c')

    .. seealso:: :class:`TaurusCommandsForm` provides a good example of use of
                 TaurusCommandButton (including managing the return value) '''
    # TODO: tango-centric
    commandExecuted = Qt.pyqtSignal(compat.PY_OBJECT)

    def __init__(self, parent=None, designMode=False, command=None,
                 parameters=None, icon=None, text=None,
                 timeout=None):
        '''Constructor

        :param parent: (Qt.QWidget or None) parent of this widget
        :param designMode: (bool) flag for Qt designer
        :param command: (str) the name of the command to be executed
        :param parameters: (sequence<str>) the list of parameteres. Default value is None meaning no parameters
        :param icon: (Qt.QIcon) icon for the button
        :param text: (str) the button text (if None passed, `command` is used)
        :param timeout: (float) the command timeout (in seconds)
        '''
        name = self.__class__.__name__
        if command is None:
            command = ""
        if parameters is None:
            parameters = []
        self._command = command
        self._parameters = parameters
        self._timeout = timeout
        self._customText = text
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        if icon is not None:
            self.setIcon(Qt.QIcon(icon))
        self.setCustomText(text)
        self.setDefault(False)
        self.setAutoDefault(False)
        self.clicked.connect(self.onClicked)

    def getDisplayValue(self):
        '''see :meth:`TaurusBaseComponent.displayValue`'''
        if self._customText is not None:
            return self._customText
        if len(self._command) == 0:
            return '---'
        modelobj = self.getModelObj()
        if modelobj is None or not hasattr(modelobj, self._command):
            return '---'
        return self._command

    @ProtectTaurusMessageBox(title="Unexpected error when executing command")
    def onClicked(self, value=0):
        return self._onClicked()

    def _onClicked(self):
        '''Slot called when the button is clicked. It executes the command with
        parameters. It may issue a warning if the command is flagged as
        dangerous. On successful execution, it returns the command result and it
        emits a "commandExecuted" signal with the result as well.

        :return: The result of the command. The type depends on the command. It
                 may be None.

        .. seealso:: :meth:`setCommand`, :meth:`setParameters`, :meth:`TaurusBaseComponent.isDangerous`
        '''

        self.debug("launch command %s" % str(self._command))
        if len(self._command) == 0:
            return
        modelobj = self.getModelObj()
        if modelobj is None or not hasattr(modelobj, self._command):
            self.warning('Device %s does not implement command %s' %
                         (modelobj, self._command))
            return

        if self.isDangerous() and not self.getForceDangerousOperations():
            result = Qt.QMessageBox.question(self, "Potentially dangerous action",
                                             "%s\nProceed?" % self.getDangerMessage(),
                                             Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel,
                                             Qt.QMessageBox.Ok)
            if result != Qt.QMessageBox.Ok:
                return
        # After all the checks, we can finally do the action
        orig_timeout = modelobj.get_timeout_millis()
        try:
            if self._timeout is not None:
                modelobj.set_timeout_millis(int(self._timeout * 1000))
            result = modelobj.command_inout(self._command, self._castParameters(
                self._parameters, self._command, modelobj))
        except Exception as e:
            self.error('Unexpected error when executing command %s of %s: %s' % (
                self._command, modelobj.getNormalName(), str(e)))
            raise
        finally:
            modelobj.set_timeout_millis(orig_timeout)

        self.commandExecuted.emit(result)
        return result

    def _castParameters(self, parameters=None, command=None, dev=None):
        '''Internal method used to cast the command paramters to the appropriate
        type required for the given command

        :param parameters: (sequence) a sequence of parameters. If None is
                           passed, the currently set parameters are used.
        :param command: (str) the command name. If None is passed, the currently
                        set command is used.
        :param dev: (taurus.core.taurusdevice.TaurusDevice) the device on which the command is
                    executed. If None is passed, the current model is used.

        :return: (sequence or scalar) a sequence of parameters (or a scalar if only one parameter)
        '''
        import PyTango

        if parameters is None:
            parameters = self._parameters
        if command is None:
            command = self._command
        if dev is None:
            dev = self.getModelObj()

        try:
            param_type = dev.command_query(command).in_type
        except Exception as e:
            self.warning(
                'Cannot get parameters info for command %s:%s' % (command, str(e)))
            return parameters
        if param_type == PyTango.CmdArgType.DevVoid:
            return None
        if PyTango.is_int_type(param_type, True):
            cast_type = int
        elif PyTango.is_float_type(param_type, True):
            cast_type = float
        elif param_type == PyTango.CmdArgType.DevVarStringArray or param_type == PyTango.CmdArgType.DevString:
            cast_type = str
        elif param_type == PyTango.CmdArgType.DevVarBooleanArray or param_type == PyTango.CmdArgType.DevBoolean:
            cast_type = bool
        else:
            self.info(
                'Unsupported parameters type (%s). Casting to "str"' % str(param_type))
            cast_type = str
        if PyTango.is_scalar_type(param_type):
            if parameters:
                return cast_type(parameters[0])
            else:
                return parameters
        else:
            return list(map(cast_type, parameters))

    def setCommand(self, commandName):
        '''sets the command to be executed when the button is clicked

        :param commandName: (str or None) the command name
        '''
        if commandName is None:
            self._command = ""
        else:
            self._command = str(commandName)
        self._setText(self.getDisplayValue())

    def getCommand(self):
        '''returns the command name to be executed when the button is clicked

        :return: (str or None) the command name
        '''
        return self._command

    def resetCommand(self):
        '''equivalent to self.setCommand(None)'''
        self.setCommand("")

    def setParameters(self, parameters):
        '''
        Sets the parameters to be used on command execution.

        :param parameters: (sequence) a sequence of parameters. If the
                           elements of the sequence are not of the right type
                           required for the parameter, an automatic conversion
                           will be attempted on execution time. As a special
                           case, if parameters is a string not starting and
                           ending in quote characters, it will be splitted on
                           whitespace to obtain a sequence of parameters. If
                           it is a string starting and ending with quotes, the
                           quotes will be removed and the quoted text will not
                           be splitted.
        '''
        if isinstance(parameters, string_types):
            parameters = str(parameters).strip()
            if parameters[0] in ('"', "'") and parameters[0] == parameters[-1]:
                parameters = [parameters[1:-1]]
            else:
                parameters = parameters.split()
        self._parameters = parameters

    def getParameters(self):
        '''returns the parameters to be used on command execution

        :param parameters: (sequence)
        '''
        return self._parameters

    def resetParameters(self):
        '''Equivalent to setParameters([])
        '''
        self.setParameters([])

    def setCustomText(self, customText=None):
        '''Sets a custom text for the button (by default it is the command name)

        :param customText: (str or None) the custom text. If None passed, it
                           will use the command name
        '''
        self._customText = customText
        self._setText(self.getDisplayValue())

    def getCustomText(self):
        '''Returns the custom text of the buttom, or None if no custom text is
        used
        '''
        return self._customText

    def resetCustomText(self):
        '''Equivalent to setCustomText(None)'''
        self.setCustomText(None)

    @Qt.pyqtSlot(float)
    @Qt.pyqtSlot(int)
    def setTimeout(self, timeout):
        '''Sets the number of seconds to wait for the result of the command.

        .. seealso:: :meth:`PyTango.DeviceProxy.command_inout`

        :param timeout: (float) the command timeout in seconds
                        (timeout <0 or timeout=None disables the timeout)
        '''
        if timeout < 0:
            timeout = None
        self._timeout = timeout

    def getTimeout(self):
        '''
        Returns the number of seconds to wait for the result of the command
        (or -1 if timeout is disabled)
        '''
        ret = self._timeout
        if ret is None or ret < 0:
            ret = -1
        return ret

    def resetTimeout(self):
        '''Equivalent to setTimeout(None)'''
        self.setTimeout(None)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'group': 'Taurus Buttons',
            'icon': 'designer:pushbutton.png',
            'module': 'taurus.qt.qtgui.button',
            'container': False}

    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    #                      Qt Properties                        #
    ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

    Model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel, TaurusBaseWidget.resetModel)

    UseParentModel = Qt.pyqtProperty("bool", TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel, TaurusBaseWidget.resetUseParentModel)

    Command = Qt.pyqtProperty("QString", getCommand, setCommand, resetCommand)

    Parameters = Qt.pyqtProperty(
        "QStringList", getParameters, setParameters, resetParameters)

    DangerMessage = Qt.pyqtProperty("QString", TaurusBaseWidget.getDangerMessage,
                                    TaurusBaseWidget.setDangerMessage, TaurusBaseWidget.resetDangerMessage)

    CustomText = Qt.pyqtProperty(
        "QString", getCustomText, setCustomText, resetCustomText)

    Timeout = Qt.pyqtProperty("double", getTimeout, setTimeout, resetTimeout)


class TaurusLockButton(Qt.QPushButton, TaurusBaseWidget):

    _LOCK_MAP = {LockStatus.Unlocked: "extra_icons:lock_unlocked.svg",
                 LockStatus.Locked: "extra_icons:lock_locked_unpreviledged.svg",
                 LockStatus.LockedMaster: "extra_icons:lock_locked.svg",
                 LockStatus.Unknown: "extra_icons:lock_unknown.svg"}

    def __init__(self, parent=None, designMode=False):
        self._lock_info = TaurusLockInfo()
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self.toggled.connect(self.on_toggle)
        self.setCheckable(True)
        self.setAutoTooltip(False)
        self.insertEventFilter(eventfilters.IGNORE_ALL)
        self.update_button()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'group': 'Taurus Buttons',
            'icon': 'designer:pushbutton.png',
            'module': 'taurus.qt.qtgui.button',
            'container': False}

    def getModelClass(self):
        return TaurusDevice

    def setModel(self, model):
        TaurusBaseWidget.setModel(self, model)
        self.update_button()

    def get_lock_info(self, cache=False):
        dev = self.getModelObj()
        if dev is not None:
            self._lock_info = dev.getLockInfo(cache=cache)
        return self._lock_info

    def update_button(self, lock_info=None):
        if lock_info is None:
            lock_info = self.get_lock_info()
        status = lock_info.status
        self.setIcon(Qt.QIcon(self._LOCK_MAP[status]))
        self.setDown(status in (LockStatus.Locked, LockStatus.LockedMaster))
        self.setToolTip(lock_info.status_msg)
        self.update()
        return lock_info

    def _on_toggle(self, down):
        dev = self.getModelObj()
        if down:
            dev.lock()
        else:
            dev.unlock()
        self.update_button()

    def on_toggle(self, down):
        try:
            self._on_toggle(down)
        except:
            import sys
            from taurus.qt.qtgui.dialog import TaurusMessageBox
            msgbox = TaurusMessageBox(*sys.exc_info())
            msgbox.setWindowTitle("Error locking device")
            if self.update_button().status == LockStatus.Locked:
                msgbox.setText(self._lock_info.status_msg)
            msgbox.exec_()

    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel, setModel,
                            TaurusBaseWidget.resetModel)


def lockButtonMain():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <full_attribute_name(s)>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus lock button demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        w = demo()
    else:
        models = list(map(str.lower, args))

        w = Qt.QWidget()
        layout = Qt.QGridLayout()
        w.setLayout(layout)
        for model in models:
            lock_button = TaurusLockButton()
            lock_button.model = model
            layout.addWidget(lock_button)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w


def commandButtonMain():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(cmd_line_parser=None)
    form = TaurusCommandButton(parent=None, designMode=False, command='DevBoolean', parameters=[
                               123], icon='logos:taurus.png', text='launch: DevBoolean 123')
    form.setModel('sys/tg_test/1')
    form.setDangerMessage(
        'Booo scary command!!\n Maybe you should think twice!')

    def f(*a):
        print(a)
    form.commandExecuted.connect(f)
    form.show()
    sys.exit(app.exec_())


def launcherButtonMain():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(cmd_line_parser=None)

    # Creating button giving the widget
    # from taurus.qt.qtgui.plot import TaurusPlot
    # w = TaurusPlot()
    # form = TaurusLauncherButton(parent=None, designMode=False, widget=w,
    #                             icon='logos:taurus.png'), text='show')

    # Creating button giving the widget class name
    # form = TaurusLauncherButton(parent=None, designMode=False,
    #                             widget='TaurusPlot', icon='logos:taurus.png',
    #                             text='show')

    # Creating button using a derived class with the name widget class
    # hardcoded
    class MyButton(TaurusLauncherButton):
        _widgetClassName = 'TaurusPlot'
        _icon = 'logos:taurus.png'
        _text = 'show'
    form = MyButton()

    form.setModel('sys/tg_test/1/wave')
    form.show()
    sys.exit(app.exec_())


def main():
    lockButtonMain()


def demo():
    '''Lock button'''
    lock_button = TaurusLockButton()
    lock_button.model = "sys/tg_test/1"
    return lock_button


if __name__ == '__main__':
    # lockButtonMain()
    launcherButtonMain()
    # commandButtonMain()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.qt.qtgui.panel.taurusvalue"""

import pytest
from taurus.qt.qtgui.panel import TaurusValue
from taurus.test.pytest import check_taurus_deprecations

try:
    # The following are Tango-centric imports.
    from taurus.core.tango.test import taurus_test_ds  # pytest fixture

    _TANGO_MISSING = False
except:
    _TANGO_MISSING = True


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
def test_bug126(qtbot, caplog):
    """Verify that case is not lost when customizing a label (bug#126)"""
    with check_taurus_deprecations(caplog, expected=0):
        w = TaurusValue()
        qtbot.addWidget(w)
        w.setModel("tango:sys/tg_test/1/double_scalar")
        label = "MIXEDcase"
        w.setLabelConfig(label)

        def _ok():
            assert w.labelWidget().text() == label

        qtbot.waitUntil(_ok, timeout=3200)


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
def test_label_case_sensitivity(qtbot, caplog, taurus_test_ds):
    """Verify that case is respected of in the label widget"""
    with check_taurus_deprecations(caplog, expected=0):
        w = TaurusValue()
        qtbot.addWidget(w)
        w.setModel("tango:{}/MIXEDcase".format(taurus_test_ds))

        def _ok():
                assert w.labelWidget().text() == "MIXEDcase"

        qtbot.waitUntil(_ok, timeout=3200)


def test_taurusvalue_subwidget_texts(qtbot, caplog):
    """Checks the texts for scalar attributes"""

    model = "eval:@a=taurus.core.evaluation.test.res.mymod.MyClass()/a.foo"
    expected = ("a.foo", "123", "123", "m")
    depr = 0

    with check_taurus_deprecations(caplog, expected=depr):
        w = TaurusValue()
        qtbot.addWidget(w)
        w.setModel(model)

        def _ok():
            got = (
                str(w.labelWidget().text()),
                str(w.readWidget().text()),
                str(w.writeWidget().displayText()),
                str(w.unitsWidget().text()),
            )
            assert got == expected

        qtbot.waitUntil(_ok, timeout=3200)




#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Examples on using the evaluation scheme for exposing icepap driver values
as taurus attributes
"""

from __future__ import print_function


ATTR_IPAP_POS = ( 'eval:@ipap=pyIcePAP.EthIcePAP("icepap06", port=5000)' +
                  '/float(ipap.readParameter(1,"POS"))')


def _test1():
    import taurus.core
    a = taurus.Attribute(ATTR_IPAP_POS)
    print("axis pos:", a.read().rvalue)


def _test2():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.qt.qtgui.display import TaurusLabel
    app = TaurusApplication(cmd_line_parser=None)

    tl = TaurusLabel()
    tl.setModel(ATTR_IPAP_POS)
    tl.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    _test1()
    _test2()



import unittest
from taurus.core.test.modelequality import TaurusModelEqualityTestCase, \
    testDeviceModelEquality, \
    testAttributeModelEquality

# the same device models, just different cases
dev_models1 = ['tango:sys/tg_test/1',
               'tango:SYS/TG_TEST/1']
# different device models
dev_models2 = ['tango:sys/tg_test/1',
               'tango:sys/database/2']
# the same attribute models, just different cases
attr_models1 = ['tango:sys/tg_test/1/float_scalar',
                'tango:SYS/TG_TEST/1/FLOAT_SCALAR']
# different attribute models (same attr from different device)
attr_models2 = ['tango:sys/tg_test/1/state',
                'tango:sys/database/2/state']
# the same attribute models, just different cases
attr_models3 = [
    'tango:sys/tg_test/1/float_scalar#label',
    'tango:SYS/TG_TEST/1/FLOAT_SCALAR#LABEL'
]
# different attribute models (same attr and fragment but different device)
attr_models4 = [
    'tango:sys/tg_test/1/state#label',
    'tango:sys/database/2/state#label'
]
# the same attribute models, just different fragment
attr_models5 = [
    'tango:sys/tg_test/1/float_scalar#label',
    'tango:SYS/TG_TEST/1/FLOAT_SCALAR#range'
]

# # test cases for the model case insensitiveness


@testAttributeModelEquality(models=attr_models1, equal=True)
@testDeviceModelEquality(models=dev_models1, equal=True)
@testAttributeModelEquality(models=attr_models3, equal=True)
# test cases for the different models inequality
@testAttributeModelEquality(models=attr_models2, equal=False)
@testDeviceModelEquality(models=dev_models2, equal=False)
@testAttributeModelEquality(models=attr_models4, equal=False)
# test cases for the different models equality
# TODO: add device/alias equality tests
@testAttributeModelEquality(models=attr_models5, equal=True)
class TangoModelEqualityTestCase(TaurusModelEqualityTestCase,
                                 unittest.TestCase):
    """TestCase class for tango model equality testing."""
    pass



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides LazyModule for loading plugin modules in lazy way -
only when members of actual plugin are requested.
"""

__all__ = ["LazyModule"]

from types import ModuleType
import sys

from taurus import warning, info


class LazyModule(ModuleType):
    """It provides a ModuleType object that acts as a placeholder for modules
    registered via entry-points. It replaces the actual module without actually
    importing it until a member of the module is requested, which automatically
    triggers the load of the entry-point and the substitution of this
    placeholder by the actual module
    """
    def __init__(self, name, package, entry_point):
        """

        :param name: name of the module
        :param package: name of the package to which the module belongs
        :param entry_point: entry-point for the module
        """
        super(LazyModule, self).__init__(name)
        self.__package__ = package
        self.ep = entry_point

    def __getattr__(self, member):
        try:
            mod = self.ep.load()
            # Replace lazy module with actual module for package
            setattr(sys.modules[self.__package__], self.__name__, mod)
            # Replace lazy module with actual module in sys.modules
            modname = "%s.%s" % (self.__package__, self.__name__)
            sys.modules[modname] = mod
        except Exception as e:
            warning('Could not load plugin "%s". Reason: %s', self.ep.name, e)
            return None
        return getattr(mod, member)

    @staticmethod
    def import_ep(name, package, entry_point):
        """
        Lazily imports a module defined in an entry point. The LazyModule is
        inserted in sys.modules as <package>.<name>

        :param name: name of the module
        :param package: name of the package to which the module belongs
        :param entry_point: entry-point for a module
        :return: A LazyModule object
        """
        m = LazyModule(name, package, entry_point)
        setattr(sys.modules[package], name, m)
        modname = "{}.{}".format(package, name)
        sys.modules[modname] = m
        info(
            'Plugin "%s" lazy-loaded as "%s"',
            entry_point.module_name,
            modname
        )
        return m



from taurus.qt.qtgui.taurusgui import TaurusGui, utils
from taurus.external.qt import PYSIDE2
import pytest

p1 = utils.PanelDescription(
    "testpanel1",
    classname="taurus.qt.qtgui.panel.TaurusForm",
    model=["eval:1"],
    widget_properties={
        "withButtons": False,
        "foobar": 34
    }
)

@pytest.mark.xfail(PYSIDE2, reason="This test is known to fail with PySide2")
def test_paneldescription(qtbot):
    gui = TaurusGui(confname=__file__, configRecursionDepth=0)
    w1 = gui.getPanel('testpanel1').widget()
    qtbot.addWidget(gui)
    qtbot.addWidget(w1)
    assert w1.withButtons is False
    assert w1.isWithButtons() is False
    assert not hasattr(w1, "foobar")
    assert w1.modifiableByUser is False


#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
TaurusPlotConf: widget for configurating the contents and appearance of a TaurusPlot
"""
from __future__ import print_function
from __future__ import absolute_import

raise NotImplementedError('Under Construction!')

import taurus.core
from taurus.external.qt import Qt, Qwt5
from taurus.qt.qtgui.util.ui import UILoadable

from . import curveprops
try:
    import taurus.qt.qtgui.extra_nexus as extra_nexus
except:
    extra_nexus = None


__all__ = ['TaurusPlotConfDlg']


@UILoadable(with_ui='ui')
class TaurusPlotConfDlg(Qt.QWidget):
    ''' A configuration dialog for TaurusPlot.

    The dialog uses a Model/View design:
      - it uses a :class:`CurvesTableModel` for describing the curves
        configuration
      - it has two views on that model: a :class:`QTableView` and a
        :class:`CurvePropertiesView`.
      - The selection is managed via a :class:`ExtendedSelectionModel` which is
        shared by both views.

    It also has a data sources browser which is used to select the sources of
    the data to be used by the curves. Currently supported sources are: Tango
    attributes, Nexus/HDF5 datasets, and column organized ASCII data. Apart from
    this, the dialog allows to use mathematical expression and the given sources
    to assign values to the curves


    When the changes are applied (eg, when the Apply button is pressed), the
    model is used to (re)configure the plot.
    '''

    def __init__(self, parent=None, curves=None):
        super(TaurusPlotConfDlg, self).__init__(parent)
        self.loadUi()

        ###################################
        #@todo: this block should disappear once we can use taurusuic4 to include taurus widgets
        self.ui.propView = self.__replaceWidget(
            curveprops.CurvePropertiesView(), self.ui.propView)
        from taurus.qt.qtgui.panel import TaurusModelSelectorTree
        tangoTree = TaurusModelSelectorTree(parent=None,
                                            selectables=[
                                                taurus.core.taurusbasetypes.TaurusElementType.Attribute],
                                            buttonsPos=Qt.Qt.RightToolBarArea)
        self.ui.tangoTree = self.__replaceWidget(tangoTree, self.ui.tangoTree)
#        l=self.ui.propView.parent().layout()
#        self.ui.propView = curveprops.CurvePropertiesView()
#        l.insertWidget(0, self.ui.propView)
#        l=self.ui.tangoTree.parent().layout()
#        from taurus.qt.qtgui.panel import TaurusModelSelectorTree
#        self.ui.tangoTree = TaurusModelSelectorTree(parent = None,
#                                                     selectables=[taurus.core.taurusbasetypes.TaurusElementType.Attribute],
#                                                     buttonsPos=Qt.Qt.RightToolBarArea)
#        l.insertWidget(0, self.ui.tangoTree)
        ####################################

        if extra_nexus is not None:
            self.ui.nexusBrowser = self.__replaceWidget(
                extra_nexus.TaurusNeXusBrowser(), self.ui.nexusBrowser)

        self.model = curveprops.CurvesTableModel(curves)
        self.selectionModel = curveprops.ExtendedSelectionModel(self.model)
        self.ui.curvesTable.setModel(self.model)
        self.ui.propView.setModel(self.model)
        self.ui.curvesTable.setSelectionModel(self.selectionModel)
        self.ui.propView.setSelectionModel(self.selectionModel)

        # host
        host = taurus.Authority().getNormalName()
        self.ui.tangoTree.setModel(host)

        # Connections
        self.ui.applyBT.clicked.connect(self.onApply)
        self.ui.reloadBT.clicked.connect(self.onReload)
        self.ui.cancelBT.clicked.connect(self.close)
        self.ui.tangoTree.addModels.connect(self.onModelsAdded)

    def __replaceWidget(self, new, old, layout=None):
        if layout is None:
            layout = old.parent().layout()
        index = layout.indexOf(old)
        layout.removeWidget(old)
        old.setParent(None)
        layout.insertWidget(index, new)
        return new

    def onModelsAdded(self, models):
        print(models)
        nmodels = len(models)
        rowcount = self.model.rowCount()
        self.model.insertRows(rowcount, nmodels)
        for i, m in enumerate(models):
            self.model.setData(self.model.index(rowcount + i, curveprops.Y),
                               value=m)

    def onApply(self):
        print("APPLY!!! (todo)")
        curveConfs = self.model.dumpData()

        for c in curveConfs:
            print(repr(c))

    def onReload(self):
        print("RELOAD!!! (todo)")


class demo(Qt.QDialog):

    def __init__(self, parent=None, curves=None):
        super(demo, self).__init__(parent)

        if curves is None:
            curves = [curveprops.CurveConf(xsrc='', ysrc='a/b/c/d', properties=None, title="tangocurve", vis=Qwt5.QwtPlot.yLeft),
                      curveprops.CurveConf(
                xsrc='[1,2,3]', ysrc='=#2.x**2', properties=None, title="parab", vis=Qwt5.QwtPlot.yLeft)
            ]

        # create table
        self.model = curveprops.CurvesTableModel(curves)
#        self.selection = Qt.QItemSelectionModel(self.model)

        self.table = Qt.QTableView(self)
        self.table.setModel(self.model)

#        print "!!!!!!!!",self.table.selectionModel()
#        self.table.setItemDelegate(commandDelegate(self))

        # create buttons
        self.posSB = Qt.QSpinBox()
        self.newSB = Qt.QSpinBox()
        self.addBT = Qt.QPushButton(u"Add")
        self.remBT = Qt.QPushButton(u"Rem")
        self.dataBT = Qt.QPushButton(u"Data")

        # put widgets in the layout
        mainLayout = Qt.QGridLayout()
        mainLayout.addWidget(self.table, 0, 0, 1, 2)
        mainLayout.addWidget(self.posSB, 1, 0)
        mainLayout.addWidget(self.newSB, 1, 1)
        mainLayout.addWidget(self.addBT, 2, 0)
        mainLayout.addWidget(self.remBT, 2, 1)
        mainLayout.addWidget(self.dataBT, 3, 0)
        self.setLayout(mainLayout)

        # connections
        self.addBT.clicked.connect(self.onAdd)
        self.remBT.clicked.connect(self.onRem)
        self.dataBT.clicked.connect(self.onData)

        # misc
        self.table.resizeColumnsToContents()
#        self.table.setShowGrid(False)
#        self.table.setSelectionMode(self.table.ExtendedSelection)
#        self.table.setSelectionBehavior(Qt.QAbstractItemView.SelectItems)
        self.model.mimeTypes()

    def onAdd(self):
        self.model.insertRows(position=self.posSB.value(),
                              rows=self.newSB.value())

    def onRem(self):
        self.model.removeRows(position=self.posSB.value(),
                              rows=self.newSB.value())

    def onData(self):
        cmds = self.model.dumpData()
        print(self.model.curves)


def main1():
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)
    form = demo()
    form.show()
    sys.exit(app.exec_())


def main2():
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)

    curves = [curveprops.CurveConf(xsrc='', ysrc='tango://host:1000/a/b/c/d', properties=None, title="tangocurve", vis=Qwt5.QwtPlot.yLeft),
              curveprops.CurveConf(
                  xsrc='=[1,2,3]', ysrc='=#2.x**2', properties=None, title="parab", vis=Qwt5.QwtPlot.yLeft)
              ]
    form = TaurusPlotConfDlg(curves=curves)
    form.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    import sys
    main2()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides base table widget"""

__all__ = ["QBaseTableWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.qt.qtgui.model import QBaseModelWidget


class QBaseTableWidget(QBaseModelWidget):

    def tableView(self):
        return self.viewWidget()

    def createViewWidget(self, klass=None):
        if klass is None:
            klass = Qt.QTableView
        table = klass()
        table.setSortingEnabled(True)
        table.sortByColumn(0, Qt.Qt.AscendingOrder)
        table.setAlternatingRowColors(True)
        table.setSelectionBehavior(Qt.QAbstractItemView.SelectRows)
        table.setSelectionMode(Qt.QAbstractItemView.ExtendedSelection)
        return table



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Finds out user inf (currently just the logged user name) for Windows and
Posix machines. sets a USER_NAME variable containing the logged user name
defines a UNKNOWN_USER variable to which username falls back.
It also provides the getSystemUserName() function"""

__all__ = ["getSystemUserName", "USER_NAME", "UNKNOWN_USER"]

__docformat__ = "restructuredtext"

#import os
import getpass

# fallback. The caller of this module can check if username != UNKNOWN_USER
UNKNOWN_USER = 'UnknownUser'


def getSystemUserName():
    """Finds out user inf (currently just the logged user name) for Windows and
    Posix machines. sets a USER_NAME variable containing the logged user name
    defines a UNKNOWN_USER variable to which username falls back.

    :return: (str) current user name
    """
    try:
        return getpass.getuser()
    except:
        return UNKNOWN_USER
#    username=''
#    if os.name is 'posix': #this works for linux and macOSX
#        username=os.getenv('USER', UNKNOWN_USER)
#    elif os.name is 'nt': #win NT, XP... (and Vista?)
#        username=os.getenv('USERNAME', UNKNOWN_USER)
#    if username == '': username=UNKNOWN_USER
#    return username

USER_NAME = getSystemUserName()



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)
panel = TaurusWidget()
layout = Qt.QVBoxLayout()
panel.setLayout(layout)
panel.setModel('sys/taurustest/1')
w1 = TaurusLabel()
w2 = TaurusLabel()
w3 = TaurusLabel()
w1.setModel('sys/taurustest/1/state')
w2.setModel('sys/taurustest/1/position')
w3.setModel('sys/taurustest/1/simulationmode')
w1.setShowQuality(False)

layout.addWidget(w1)
layout.addWidget(w2)
layout.addWidget(w3)
panel.show()

sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This is an experimental property parser"""

from __future__ import print_function
from builtins import str

import os
import ply.lex as lex
import ply.yacc as yacc
from .containers import CaselessDict
from .log import Logger

reserved = CaselessDict({
    'true': 'true',
    'false': 'false',
    'yes': 'yes',
    'no': 'no'
})

tokens = ['EQUALS',
          'KEY',
          'STRING',
          'NUMBER',
          'COMMA',
          'LLST', 'RLST',
          #'LBRACKET', 'RBRACKET',
          ] + list(reserved.values())

t_EQUALS = r'\='
t_LLST = r'\['
t_RLST = r'\]'
t_COMMA = r'\,'
#t_LBRACKET = r'\{'
#t_RBRACKET = r'\}'


def t_NUMBER(t):
    r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?'
    try:
        if t.value.count('.') == 0:
            try:
                t.value = int(t.value)
            except:
                t.value = float(t.value)
        else:
            t.value = float(t.value)
    except:
        print("[%d]: Number %s is not valid!" % (t.lineno, t.value))
        t.value = 0
    return t

# The following backslash sequences have their usual special meaning: \",
# \\, \n, and \t.


def t_STRING(t):
    r'\"([^\\"]|(\\.))*\"'
    escaped = 0
    str = t.value[1:-1]
    new_str = ""
    for i in range(0, len(str)):
        c = str[i]
        if escaped:
            if c == "n":
                c = "\n"
            elif c == "t":
                c = "\t"
            new_str += c
            escaped = 0
        else:
            if c == "\\":
                escaped = 1
            else:
                new_str += c
    t.value = new_str
    return t


def t_KEY(t):
    r'[a-zA-Z0-9/_\.\/]+'
    t.type = reserved.get(t.value, 'KEY')    # Check for reserved words
    return t


# Ignored characters
t_ignore = ' \t'

t_ignore_COMMENT = r'\#.*'


def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")


def t_error(t):
    print("[%d]: Illegal character '%s'" % (t.lexer.lineno, t.value[0]))
    t.lexer.skip(1)


def p_error(p):
    print("[%d]: Syntax error in input [%s]" % (p.lineno, (str(p))))

#-------------------------------------------------------------------------
# Yacc Starting symbol
#-------------------------------------------------------------------------


def p_property_file(p):
    '''property_file : statement_list
                     |'''
    if len(p) == 1:
        p[0] = {}
    else:
        p[0] = p[1]


def p_statement_list(p):
    '''statement_list : statement_list statement'''
    p[1].update(p[2])
    p[0] = p[1]


def p_statement(p):
    '''statement_list : statement'''
    p[0] = p[1]


def p_statement_assign(p):
    '''statement : KEY EQUALS value'''
    p[0] = {p[1]: p[3]}


def p_value(p):
    '''value : atomic
             | list'''
    p[0] = p[1]


def p_value_atomic(p):
    '''atomic : KEY
              | STRING
              | NUMBER
              | boolean'''
    p[0] = p[1]


def p_value_boolean(p):
    '''boolean : true
               | false
               | yes
               | no'''
    v = p[1].lower()
    p[0] = v in ('true', 'yes')


def p_value_list(p):
    '''list : LLST value_list RLST'''
    p[0] = p[2]


def p_value_list_elems1(p):
    '''value_list : value_list COMMA value'''
    p[0] = p[1] + [p[3]]


def p_value_list_elems2(p):
    '''value_list : value'''
    p[0] = [p[1]]


class PropertyParser(Logger):

    def __init__(self, parent=None):
        name = self.__class__.__name__
        self.call__init__(Logger, name, parent)
        self._last_filename = None

    def parse_file(self, f, logger=None, debug=0, optimize=1):
        self._last_filename = os.path.abspath(f.name)

        kw = {
            'debuglog': logger or self,
            'errorlog': logger or self,
            'debug': debug,
            'optimize': optimize,
        }

        kw_lex = kw.copy()
        kw_lex.update({'lextab': 'property_lextab'})
        kw_yacc = kw.copy()
        kw_yacc.update({'tabmodule': 'property_yacctab'})
        lexer = lex.lex(**kw_lex)
        parser = yacc.yacc(**kw_yacc)

        if debug:
            debug = self
        return parser.parse(f.read(), debug=debug, lexer=lexer)

    def parse(self, filename, logger=None, debug=0, optimize=1):
        filename = os.path.realpath(filename)
        f, res = None, None
        try:
            f = open(filename)
            res = self.parse_file(
                f, logger=logger, debug=debug, optimize=optimize)
            f.close()
        except IOError as e:
            if f:
                f.close()
            raise
        return res

    def getLastFilename(self):
        return self._last_filename

if __name__ == '__main__':
    import sys
    pp = PropertyParser()
    res = pp.parse(sys.argv[1])
    print(res)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.tango.test.test_tangovalidator..."""

#__all__ = []

__docformat__ = 'restructuredtext'


import unittest
from taurus.core.test import (valid, invalid, names,
                              AbstractNameValidatorTestCase)
from taurus.core.tango.tangovalidator import (TangoAuthorityNameValidator,
                                              TangoDeviceNameValidator,
                                              TangoAttributeNameValidator)

import PyTango
from taurus.core.util.fqdn import fqdn_no_alias

__PY_TANGO_HOST = PyTango.ApiUtil.get_env_var("TANGO_HOST")
host, port = __PY_TANGO_HOST.split(':')
__TANGO_HOST = "{0}:{1}".format(fqdn_no_alias(host), port)

#=========================================================================
# Tests for Tango Authority  name validation
#=========================================================================
@valid(name='tango://foo:10000')
@invalid(name='tango:foo')
@invalid(name='tango:foo:10000')
@invalid(name='tango://foo:10000/')
@invalid(name='tango://foo:10000/?')
@invalid(name='tango://foo:bar')
@invalid(name='tango://foo:10000/foo')
@names(name='tango://foo:123',
       out=('tango://foo:123', '//foo:123', 'foo:123'))
@names(name='//foo:123',  # using implicit scheme!
       out=('tango://foo:123', '//foo:123', 'foo:123'))
class TangoAuthValidatorTestCase(AbstractNameValidatorTestCase,
                                 unittest.TestCase):
    validator = TangoAuthorityNameValidator


#=========================================================================
# Tests for Tango Device name validation
#=========================================================================
@valid(name='tango:foo', groups={'devname': 'foo',
                                 '_devalias': 'foo',
                                 '_devslashname': None})
@valid(name='tango:a/b/c', groups={'devname': 'a/b/c',
                                   '_devalias': None,
                                   '_devslashname': 'a/b/c'})
@valid(name='tango://foo:123/a/b/c', groups={'devname': 'a/b/c',
                                             '_devalias': None,
                                             '_devslashname': 'a/b/c'})
@valid(name='tango:a/b/ c', groups={'devname': 'a/b/ c'})
@invalid(name='tango:/a/b/c?')
@valid(name='tango://a/b/c', strict=False)
@valid(name='tango:alias')
@valid(name='tango://alias', strict=False)
@valid(name='tango://a/b/c', strict=False)
@invalid(name='tango:foo:1234/alias', strict=False)
@invalid(name='tango:foo:1234/a/b/c', strict=False)
@valid(name='foo:1234/alias', strict=False)  # Implicit scheme
@valid(name='foo:1234/a/b/c', strict=False)  # Implicit scheme
@invalid(name='tango://a/b/c', strict=True)
@invalid(name='tango://devalias')
@names(name='tango://foo:123/a/b/c',
       out=('tango://foo:123/a/b/c', '//foo:123/a/b/c', 'a/b/c'))
@names(name='tango:sys/tg_test/1',
       out=('tango://%s/sys/tg_test/1' % __TANGO_HOST,
            'sys/tg_test/1', 'sys/tg_test/1'))
@names(name='tango:alias', out=(None, None, 'alias'))
# @names(name = 'tango:mot49', # commented out because it assumes mot49 exists
#        out=('tango://%s/motor/motctrl13/1'% __TANGO_HOST,
#             'motor/motctrl13/1', 'mot49'))
class TangoDevValidatorTestCase(AbstractNameValidatorTestCase,
                                unittest.TestCase):
    validator = TangoDeviceNameValidator


#=========================================================================
# Tests for Tango Attribute name validation (without fragment)
#=========================================================================
@valid(name='foo:10000/a/b/c/d', strict=False)
@valid(name='mot/position', strict=False)
@valid(name='tango:a/b/c/d', groups={'devname': 'a/b/c',
                                     'attrname': 'a/b/c/d',
                                     '_shortattrname': 'd'})
@valid(name='tango:alias/attr', groups={'devname': 'alias',
                                        '_devalias': 'alias',
                                        'attrname': 'alias/attr',
                                        '_shortattrname': 'attr'})
@valid(name='tango://a/b/c/d', strict=False)
@valid(name='tango://alias/attr', strict=False)
@invalid(name='tango://a/b/c/d')
@invalid(name='tango://a/b/c/')
@invalid(name='tango://alias/attr')
@invalid(name='tango://a/b/c/d?')
@invalid(name='tango:a/b/c')
@invalid(name='tango:sys/tg_test/1')
@names(name='tango://foo:123/a/b/c/d',
       out=('tango://foo:123/a/b/c/d', '//foo:123/a/b/c/d', 'd'))
@names(name='tango:sys/tg_test/1/float_scalar',
       out=('tango://%s/sys/tg_test/1/float_scalar' % __TANGO_HOST,
            'sys/tg_test/1/float_scalar', 'float_scalar'))
# @names(name = 'tango:mot49/position', # commented out because it assumes mot49
#        out=('tango://%s/motor/motctrl13/1/position'% __TANGO_HOST,
#             'motor/motctrl13/1/position', 'position'))
#=========================================================================
# Tests for validation of Attribute name with fragment
#=========================================================================
@valid(name='tango:a/b/c/d#',
       groups={'devname': 'a/b/c', 'attrname': 'a/b/c/d', '_shortattrname': 'd',
               'cfgkey': ''})
@valid(name='tango:a/b/c/d#label',
       groups={'devname': 'a/b/c', 'attrname': 'a/b/c/d', '_shortattrname': 'd',
               'cfgkey': 'label'})
@valid(name='tango:alias/attr#')
@valid(name='tango:alias/attr#label')
@valid(name='tango://a/b/c/d?configuration', strict=False)
@valid(name='tango://a/b/c/d?configuration=label', strict=False,
       groups={'devname': 'a/b/c',
               'attrname': 'a/b/c/d',
               '_shortattrname': 'd',
               'cfgkey': 'label',
               'fragment': 'label'})  # fragment=cfgkey (in non-strict mode)
@invalid(name='tango://a/b/c/d#')
@invalid(name='tango://a/b/c/d?foo', strict=False)
@invalid(name='tango://a/b/c/d?foo=label', strict=False)
@valid(name='tango:a/b/c/d#?foo=label')  # "?" is allowed in cfgkeys!
@valid(name='tango:a/b/c/d#label?foo=bar')
@names(name='tango://foo:123/a/b/c/d#',
       out=('tango://foo:123/a/b/c/d',
            '//foo:123/a/b/c/d', 'd', ''))
@names(name='tango://foo:123/a/b/c/d#label',
       out=('tango://foo:123/a/b/c/d',
            '//foo:123/a/b/c/d', 'd', 'label'))
@names(name='tango:sys/tg_test/1/float_scalar#',
       out=('tango://%s/sys/tg_test/1/float_scalar' % __TANGO_HOST,
            'sys/tg_test/1/float_scalar', 'float_scalar', ''))
@names(name='tango://foo:123/a/b/c/d?configuration=label',
       out=('tango://foo:123/a/b/c/d',
            '//foo:123/a/b/c/d', 'd', 'label'))
class TangoAttrValidatorTestCase(AbstractNameValidatorTestCase,
                                 unittest.TestCase):
    validator = TangoAttributeNameValidator


if __name__ == '__main__':
    pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the `TangoFactory` object"""

from __future__ import absolute_import
from future.utils import string_types

try:
    pass
except ImportError:

    # note that if PyTango is not installed the factory will not be available
    from taurus.core.util.log import debug
    msg = 'cannot import PyTango module. ' + \
          'Taurus will not support the "tango" scheme'
    debug(msg)
    raise
import PyTango

from taurus import tauruscustomsettings
from taurus.core.taurusbasetypes import (TaurusElementType,
                                         TaurusSerializationMode)
from taurus.core.taurusfactory import TaurusFactory
from taurus.core.taurusbasetypes import OperationMode
from taurus.core.taurusexception import TaurusException, DoubleRegistration
from taurus.core.util.log import Logger, taurus4_deprecation
from taurus.core.util.singleton import Singleton
from taurus.core.util.containers import CaselessWeakValueDict, CaselessDict

from .tangodatabase import TangoAuthority
from .tangoattribute import TangoAttribute
from .tangodevice import TangoDevice

_Authority = TangoAuthority
_Attribute = TangoAttribute
_Device = TangoDevice


__all__ = ["TangoFactory"]
__docformat__ = "restructuredtext"


class TangoFactory(Singleton, TaurusFactory, Logger):
    r"""A :class:`TaurusFactory` singleton class to provide Tango-specific
    Taurus Element objects (TangoAuthority, TangoDevice, TangoAttribute)

    Tango model names are URI based See https://tools.ietf.org/html/rfc3986.
    For example, a TangoAttribute would be::

        tango://foo.org:1234/a/b/c/d#label
        \___/   \_____/ \__/ \_____/ \___/
          |        |     |      |      |
          |    hostname port  attr     |
          |   \____________/\______/   |
          |         |           |      |
        scheme   authority     path  fragment

    For Tango Elements:

        - The 'scheme' must be the string "tango" (lowercase mandatory)
        - The 'authority' identifies the Tango database (<hostname> and <port>
          are mandatory if authority is given)
        - The 'path' identifies Tango Device and Attributes.
          For devices it must have the format _/_/_ or alias
          For attributes it must have the format _/_/_/_ or devalias/_
        - The 'fragment' is optional and it refers to a member of the model
          object, thus not being part of the model name itself
    """

    #: the list of schemes that this factory supports. For this factory: 'tango'
    #: is the only scheme
    schemes = ("tango",)
    caseSensitive = False
    elementTypesMap = {TaurusElementType.Authority: TangoAuthority,
                       TaurusElementType.Device: TangoDevice,
                       TaurusElementType.Attribute: TangoAttribute
                       }

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args, **kwargs):
        """Singleton instance initialization.
           **For internal usage only**"""
        name = self.__class__.__name__
        self.call__init__(Logger, name)
        self.call__init__(TaurusFactory)
        self._polling_enabled = True
        self.reInit()
        self.scheme = 'tango'
        self._serialization_mode = TaurusSerializationMode.get(
            getattr(tauruscustomsettings, 'TANGO_SERIALIZATION_MODE',
                    'TangoSerial'))

    def reInit(self):
        """Reinitialize the singleton"""
        self._default_tango_host = None
        self._tango_subscribe_enabled = True
        self.dft_db = None
        self.tango_db = CaselessWeakValueDict()
        self.tango_db_queries = CaselessWeakValueDict()
        self.tango_attrs = CaselessWeakValueDict()
        self.tango_devs = CaselessWeakValueDict()
        self.tango_dev_queries = CaselessWeakValueDict()
        self.tango_alias_devs = CaselessWeakValueDict()
        self.polling_timers = {}

        # Plugin device classes
        self.tango_dev_klasses = {}

        # Plugin attribute classes
        self.tango_attr_klasses = CaselessDict()

    def cleanUp(self):
        """Cleanup the singleton instance"""
        self.trace("[TangoFactory] cleanUp")
        for k, v in self.tango_attrs.items():
            v.cleanUp()
        for k, v in self.tango_dev_queries.items():
            v.cleanUp()
        for k, v in self.tango_devs.items():
            v.cleanUp()
        self.dft_db = None
        for k, v in self.tango_db_queries.items():
            v.cleanUp()
        for k, v in self.tango_db.items():
            v.cleanUp()
        try:
            util = PyTango.Util.instance(False)
        except PyTango.DevFailed:
            PyTango.ApiUtil.cleanup()
        self.reInit()

    def getExistingAttributes(self):
        """Returns a new dictionary will all registered attributes on this factory

           :return:  dictionary will all registered attributes on this factory
           :rtype: dict"""
        return dict(self.tango_attrs)

    def getExistingDevices(self):
        """Returns a new dictionary will all registered devices on this factory

           :return:  dictionary will all registered devices on this factory
           :rtype: dict"""
        return dict(self.tango_devs)

    def getExistingDatabases(self):
        """Returns a new dictionary will all registered databases on this factory

           :return:  dictionary will all registered databases on this factory
           :rtype: dict"""
        return dict(self.tango_db)

    def set_default_tango_host(self, tango_host):
        """
        Sets the new default tango host. The method will transform the given
        name to an Authority URI.
        
        .. note:: Calling this method also clears the device alias cache.

        :param tango_host: (str) the new tango host. It accepts any valid Tango
                                 authority name or None to use the defined by
                                 $TANGO_HOST env. var.

        """
        # Translate to Authority URI
        if tango_host and "//" not in tango_host:
            tango_host = "//{0}".format(tango_host)
        v = self.getAuthorityNameValidator()
        self._default_tango_host = v.getUriGroups(tango_host)["authority"]
        self.tango_alias_devs.clear()
        self.dft_db = None

    def get_default_tango_host(self):
        """Retruns the current default tango host
        """
        return self._default_tango_host
    
    def set_tango_subscribe_enabled(self,value):
        """ If True, enables event subscribing on TangoAttribute objects

        .. warning:: This method belongs to a "Delayed Event Subscription" API
                 added in v.4.2.1-alpha as an *experimental* feature. This API
                 may not be stable and/or it may be removed in a future release
                 (even on a minor version change)
        """
        self._tango_subscribe_enabled = value
    
    def is_tango_subscribe_enabled(self):
        """ Returns the current tango_subscribe_enabled status

        .. warning:: This method belongs to a "Delayed Event Subscription" API
                 added in v.4.2.1-alpha as an *experimental* feature. This API
                 may not be stable and/or it may be removed in a future release
                 (even on a minor version change)
        """
        return self._tango_subscribe_enabled

    def registerAttributeClass(self, attr_name, attr_klass):
        """Registers a new attribute class for the attribute name.

           :param attr_name: (str) attribute name
           :param attr_klass: (taurus.core.tango.TangoAttribute) the new class that
                              will handle the attribute
        """
        self.tango_attr_klasses[attr_name] = attr_klass

    def unregisterAttributeClass(self, attr_name):
        """Unregisters the attribute class for the given attribute
           If no class was registered before for the given attribute, this call
           as no effect

           :param attr_name: (str) attribute name
        """
        if attr_name in self.tango_attr_klasses:
            del self.tango_attr_klasses[attr_name]

    def registerDeviceClass(self, dev_klass_name, dev_klass):
        """Registers a new python class to handle tango devices of the given tango class name

           :param dev_klass_name: (str) tango device class name
           :param dev_klass: (taurus.core.tango.TangoDevice) the new class that will
                             handle devices of the given tango class name
        """
        self.tango_dev_klasses[dev_klass_name] = dev_klass

    def unregisterDeviceClass(self, dev_klass_name):
        """Unregisters the class for the given tango class name
           If no class was registered before for the given attribute, this call
           as no effect

           :param dev_klass_name: (str) tango device class name
        """
        if dev_klass_name in self.tango_dev_klasses:
            del self.tango_dev_klasses[dev_klass_name]

    def getDatabase(self, name=None):
        '''Deprecated. Use getAuthority instead'''
        return self.getAuthority(name=name)

    def getAuthority(self, name=None):
        """
        Obtain the object corresponding to the given database name or the
        default database if name is None.
        If the corresponding authority object already exists, the existing
        instance is returned. Otherwise a new instance is stored and returned.

        :param name: (str) database name string alias. If None, the
                        default database is used

        :return: (taurus.core.tangodatabase.TangoAuthority) database object
        :raise: (taurus.core.taurusexception.TaurusException) if the given alias is invalid.
        """
        ret = None
        if name is None:
            if self.dft_db is None:
                try:
                    if self._default_tango_host is None:
                        self.dft_db = _Authority()
                    else:
                        name = self._default_tango_host
                        validator = _Authority.getNameValidator()
                        groups = validator.getUriGroups(name)
                        if groups is None:
                            raise TaurusException(
                                "Invalid default Tango authority name %s" % name)
                        self.dft_db = _Authority(host=groups['host'],
                                                 port=groups['port'])
                except:
                    self.debug("Could not create Authority", exc_info=1)
                    raise
                name = self.dft_db.getFullName()
                self.tango_db[name] = self.dft_db
            ret = self.dft_db
        else:
            ret = self.tango_db.get(name)
            if not ret is None:
                return ret
            validator = _Authority.getNameValidator()
            groups = validator.getUriGroups(name)
            if not validator.isValid(name):
                raise TaurusException("Invalid Tango authority name %s" % name)
            try:
                ret = _Authority(host=groups['host'], port=groups['port'])
            except:
                self.debug("Could not create Authority %s", groups['authority'],
                           exc_info=1)

            if ret is not None:
                self.tango_db[name] = ret
        return ret

    def getDevice(self, dev_name, create_if_needed=True, **kw):
        """Obtain the object corresponding to the given tango device name.
           If the corresponding device already exists, the existing instance
           is returned. Otherwise a new instance is stored and returned.

           :param dev_name: (str) tango device name or tango alias for the
                            device. It must be a valid Tango device URI.
                            If authority is not explicit, the default Tango
                            Database will be used
           :param create_if_needed: (bool) If True, the Device is created if it
                                    did not exist previously. If False, it
                                    returns None if it did not exist

           :return: (taurus.core.tango.TangoDevice) a device object
           :raise: (taurus.core.taurusexception.TaurusException) if the given
                   dev_name is invalid.
        """
        d = self.tango_devs.get(dev_name)
        if d is None:
            d = self.tango_alias_devs.get(dev_name)
        if d is not None:
            return d

        validator = _Device.getNameValidator()
        groups = validator.getUriGroups(dev_name)

        if groups is None:
            raise TaurusException("Invalid Tango device name '%s'" % dev_name)

        full_dev_name, _, _ = validator.getNames(dev_name)

        if full_dev_name is None:
            raise TaurusException("Cannot find full name of '%s'" % dev_name)

        d = self.tango_devs.get(full_dev_name)

        if not create_if_needed:
            return d

        if d is None:
            try:
                db = self.getAuthority(groups.get('authority'))
                dev_klass = self._getDeviceClass(
                    db=db, devname=groups['devname'])
                kw['storeCallback'] = self._storeDevice
                kw['parent'] = db
                d = dev_klass(full_dev_name, **kw)
                # device objects will register themselves in this factory
                # so there is no need to do it here
            except DoubleRegistration:
                d = self.tango_devs.get(full_dev_name)
            except:
                self.debug("Error creating device %s", dev_name, exc_info=1)
                raise
        return d

    def getAttribute(self, attr_name, create_if_needed=True, **kwargs):
        """Obtain the object corresponding to the given attribute name.
           If the corresponding attribute already exists, the existing instance
           is returned. Otherwise a new instance is stored and returned.

           :param attr_name: (str) a valid attribute name URI
           :param create_if_needed: (bool) If True, the Attribute is created if
                                    it did not already exist. If False,
                                    None is returned if it did not exist
           :return: (taurus.core.tangoattribute.TangoAttribute) attribute object
           :raise: (taurus.core.taurusexception.TaurusException) if the given
                   alias is invalid.
        """
        attr = self.tango_attrs.get(attr_name)
        if attr is not None:
            return attr

        # Simple approach did not work. Lets build a proper device name
        validator = _Attribute.getNameValidator()
        groups = validator.getUriGroups(attr_name)
        if groups is None:
            raise TaurusException(("Invalid Tango attribute name '%s'") %
                                  attr_name)

        full_attr_name, _, _ = validator.getNames(attr_name)

        if full_attr_name is None:
            raise TaurusException("Cannot find full name of '%s'" % attr_name)

        attr = self.tango_attrs.get(full_attr_name)

        if attr is None:
            dev_name = full_attr_name.rsplit('/', 1)[0]
            try:
                dev = self.getDevice(dev_name)
                if dev is not None:
                    # Do another try in case the Device object created the attribute
                    # itself. This happens for the 'state' attribute
                    attr = self.tango_attrs.get(full_attr_name)
                    if attr is not None:
                        return attr
                    try:
                        attr_klass = self._getAttributeClass(
                            attr_name=attr_name)
                        kwargs['storeCallback'] = self._storeAttribute
                        if 'pollingPeriod' not in kwargs:
                            kwargs[
                                'pollingPeriod'] = self.getDefaultPollingPeriod()
                        attr = attr_klass(full_attr_name, dev, **kwargs)
                        # attribute objects will register themselves in this factory
                        # so there is no need to do it here
                    except DoubleRegistration:
                        attr = self.tango_attrs.get(full_attr_name)
            except:
                self.debug("Error creating attribute %s",
                           attr_name, exc_info=1)
                raise
        return attr

    def getAttributeInfo(self, full_attr_name):
        """Deprecated: Use :meth:`taurus.core.tango.TangoFactory.getConfiguration` instead.

           Obtain attribute information corresponding to the given attribute name.
           If the corresponding attribute info already exists, the existing information
           is returned. Otherwise a new information instance is stored and returned.

           :param full_attr_name: (str) attribute name in format: <tango device name>'/'<attribute name>

           :return: (taurus.core.tango.TangoConfiguration) configuration object
        """
        self.deprecated("Use getConfiguration(full_attr_name) instead")
        attr = self.getAttribute(full_attr_name)
        return attr

    @taurus4_deprecation(alt='getAttribute')
    def getConfiguration(self, param):
        """Obtain the object corresponding to the given attribute or full name.
           If the corresponding configuration already exists, the existing instance
           is returned. Otherwise a new instance is stored and returned.

           :param param: (taurus.core.taurusattribute.TaurusAttribute or str)
                         attribute object or full configuration name

           :return: (taurus.core.tango.TangoAttribute) configuration object
        """
        if isinstance(param, string_types):
            return self.getAttribute(param)
        return param

    def _getAttributeClass(self, **params):
        attr_name = params.get("attr_name")
        attr_klass = self.tango_attr_klasses.get(attr_name, _Attribute)
        return attr_klass

    def _getDeviceClass(self, **kwargs):
        db, dev_name = kwargs.get("db"), kwargs.get("devname")
        if db is None or dev_name is None or len(self.tango_dev_klasses) == 0:
            return _Device
        else:
            if '/' not in dev_name:  # we got an alias... find the devslashname
                dev_name = db.getElementFullName(dev_name)
            try:
                tango_dev_klass = db.get_class_for_device(dev_name)
            except PyTango.DevFailed:
                # sometimes we can't get the class (e.g. dev_name not defined)
                return _Device
            return self.tango_dev_klasses.get(tango_dev_klass, _Device)

    def _storeDevice(self, dev):
        name, alias = dev.getFullName(), dev.getSimpleName()
        exists = self.tango_devs.get(name)
        if not exists is None:
            if exists == dev:
                msg = "%s has already been registered before" % name
            else:
                msg = "%s has already been registered before with a different object!" % name
            self.debug(msg)
            raise DoubleRegistration(msg)
        self.tango_devs[name] = dev
        if not alias is None and len(alias):
            self.tango_alias_devs[alias] = dev

    def _storeAttribute(self, attr):
        name = attr.getFullName()
        exists = self.tango_attrs.get(name)
        if not exists is None:
            if exists == attr:
                msg = "%s has already been registered before" % name
            else:
                msg = "%s has already been registered before with a different object!" % name
            self.debug(msg)
            raise DoubleRegistration(msg)
        self.tango_attrs[name] = attr

    def getExistingAttribute(self, attr_name):
        """Deprecated: use getAtribute with create_if_needed=False
        """
        self.warning(('getExistingAttribute is deprecated. ' +
                      'Use getDevice with create_if_needed=False'))
        return self.getAttribute(attr_name, create_if_needed=False)

    def getExistingDevice(self, dev_name):
        """Deprecated: use getDevice with create_if_needed=False
        """
        self.warning(('getExistingDevice is deprecated. ' +
                      'Use getDevice with create_if_needed=False'))
        return self.getDevice(dev_name, create_if_needed=False)

    def removeExistingDevice(self, dev_or_dev_name):
        """Removes a previously registered device.

           :param dev_or_dev_name: (str or TangoDevice) device name or device object
        """
        if isinstance(dev_or_dev_name, _Device):
            dev = dev_or_dev_name
        else:
            dev = self.getDevice(dev_or_dev_name, create_if_needed=False)
        if dev is None:
            raise KeyError("Device %s not found" % dev_or_dev_name)
        dev.cleanUp()
        full_name = dev.getFullName()
        if full_name in self.tango_devs:
            del self.tango_devs[full_name]
        simp_name = dev.getSimpleName()
        if simp_name in self.tango_alias_devs:
            del self.tango_alias_devs[simp_name]

    def removeExistingAttribute(self, attr_or_attr_name):
        """Removes a previously registered attribute.

           :param attr_or_attr_name: (str or TangoAttribute) attribute name or attribute object
        """
        if isinstance(attr_or_attr_name, _Attribute):
            attr = attr_or_attr_name
        else:
            attr = self.getExistingAttribute(attr_or_attr_name)
        if attr is None:
            raise KeyError("Attribute %s not found" % attr_or_attr_name)
        attr.cleanUp()
        full_name = attr.getFullName()
        if full_name in self.tango_attrs:
            del self.tango_attrs[full_name]

    def isPollingEnabled(self):
        """Tells if the local tango polling is enabled

           :return: (bool) wheter or not the polling is enabled
        """
        return self._polling_enabled

    def disablePolling(self):
        """Disable the application tango polling"""
        if not self.isPollingEnabled():
            return
        self._polling_enabled = False
        for period, timer in self.polling_timers.items():
            timer.stop()

    def enablePolling(self):
        """Enable the application tango polling"""
        if self.isPollingEnabled():
            return
        for period, timer in self.polling_timers.items():
            timer.start()
        self._polling_enabled = True

    def getDatabaseNameValidator(self):
        """Deprecated"""
        self.warning(('getDatabaseNameValidator is deprecated.' +
                      'Use "Authority" instead of "Database"'))
        return self.getAuthorityNameValidator()

    def getAuthorityNameValidator(self):
        """Return TangoAuthorityNameValidator"""
        from . import tangovalidator
        return tangovalidator.TangoAuthorityNameValidator()

    def getDeviceNameValidator(self):
        """Return TangoDeviceNameValidator"""
        from . import tangovalidator
        return tangovalidator.TangoDeviceNameValidator()

    def getAttributeNameValidator(self):
        """Return TangoAttributeNameValidator"""
        from . import tangovalidator
        return tangovalidator.TangoAttributeNameValidator()

    def setOperationMode(self, mode):
        """ Deprecated. setOperationMode(OperationMode mode) -> None
            Sets the operation mode for the Tango system."""
        dep = 'setOperationMode'
        rel = 'Taurus4'
        dbg_msg = "Don't use this method"
        msg = '%s is deprecated (from %s). %s' % (dep, rel, dbg_msg)
        self.deprecated(msg)

    def getOperationMode(self):
        """Deprecated. Gives the current operation mode."""
        dep = 'getOperationMode'
        rel = 'Taurus4'
        dbg_msg = "Don't use this method"
        msg = '%s is deprecated (from %s). %s' % (dep, rel, dbg_msg)
        self.deprecated(msg)
        return OperationMode.ONLINE



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)
panel = Qt.QWidget()
layout = Qt.QHBoxLayout()
panel.setLayout(layout)

from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
w1, w2, w3 = TaurusLabel(), TaurusLabel(), TaurusLabel()
layout.addWidget(w1)
layout.addWidget(w2)
layout.addWidget(w3)
w1.model, w1.bgRole = 'sys/tg_test/1/double_scalar#label', ''
w2.model = 'sys/tg_test/1/double_scalar'
w3.model, w3.bgRole = 'sys/tg_test/1/double_scalar#unit', ''

panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for Taurus Forms"""

import unittest
from taurus.qt.qtgui.test import GenericWidgetTestCase
from taurus.qt.qtgui.panel import TaurusForm, TaurusAttrForm, TaurusValue
from taurus.core.util.test.test_plugin import mock_entry_point


class _DummyTV(TaurusValue):
    pass


def _DummyItemFactory(m):
    """
    A dummy item factory that returns _DummyTV instance for one specific
    attribute: "eval://localhost/@dummy/'test_itemfactory'"
    """
    if m.fullname == "eval://localhost/@dummy/'test_itemfactory'":
        return _DummyTV()


def _BadFactory(m):
    """
    A dummy item factory that fails when called with the attribute
    "eval://localhost/@dummy/'test_badfactory'" and returns _DummyTV otherwise.
    """
    if m.fullname == "eval://localhost/@dummy/'test_badfactory'":
        return _DummyTV()
    raise RuntimeError("_BadFactory is doomed to fail")


class _BadEntryPoint(object):
    """A dummy entry point -like class that fails when loaded (for testing)"""
    name = '_BadEntryPoint'
    def load(self):
        raise RuntimeError("_BadEntryPoint is doomed to fail")


def test_form_itemFactory(qtbot):
    """Checks that the TaurusForm itemFactory API works"""
    lines = ["test_Form_ItemFactory={}:_DummyItemFactory".format(__name__)]
    group = "taurus.form.item_factories"
    mock_entry_point(lines, group=group)

    w = TaurusForm()
    qtbot.addWidget(w)

    w.setModel(
        [
            "eval://localhost/@dummy/'test_itemfactory'",
            "eval://localhost/@dummy/'test_itemfactory2'",
        ]
    )
    qtbot.wait_until(lambda: len(w) == 2, timeout=3200)
    # The first item should get a customized _DummyTV widget
    assert type(w[0]) is _DummyTV
    # The second item shoud get the default form widget
    assert type(w[1]) is w._defaultFormWidget


def test_form_itemFactory_selection(qtbot):
    """Checks that the TaurusForm itemFactory selection API works"""
    lines = ["test_Form_ItemFactorySel={}:_DummyItemFactory".format(__name__)]
    group = "taurus.form.item_factories"
    mapping = mock_entry_point(lines, group=group)
    ep1 = mapping[group]["test_Form_ItemFactorySel"]

    w = TaurusForm()
    qtbot.addWidget(w)

    # the test_Form_ItemFactory should be in the default factories
    default_factories = w.getItemFactories()
    assert ep1 in default_factories

    # no factories should be excluded by default
    inc, exc = w.getItemFactories(return_disabled=True)
    assert exc == []

    # Check that we can deselect all factories
    no_factories = w.setItemFactories(include=[])
    assert no_factories == []

    # Check that we can exclude everything except test_Form_ItemFactory
    select1 = w.setItemFactories(
        exclude=[r"(?!.*test_Form_ItemFactorySel).*"]
    )
    assert select1 == [ep1]

    # Check that we can include only test_Form_ItemFactory
    select2 = w.setItemFactories(include=["test_Form_ItemFactorySel"])
    assert select2 == [ep1]

    # Check that the selected test_Form_ItemFactory is an entry point
    from pkg_resources import EntryPoint
    assert type(select2[0]) == EntryPoint

    # Check that the selected entry point loads _DummyItemFactory
    assert select2[0].load() is _DummyItemFactory

    # Check that we can include a factory instance
    select3 = w.setItemFactories(include=[_DummyItemFactory])

    # Check that the selected test_Form_ItemFactory is an entry point-alike
    from taurus.core.util.plugin import EntryPointAlike
    assert type(select3[0]) == EntryPointAlike

    # Check that the selected entry point loads _DummyItemFactory
    assert select3[0].load() is _DummyItemFactory

    # Check that the selected entry point has the given name
    assert select3[0].name == repr(_DummyItemFactory)


def test_form_cwidget_bck_compat(qtbot):
    """check that the cusomWidgetMap bck-compat works"""

    w = TaurusForm()
    qtbot.addWidget(w)

    # check that custom widget map is empty by default
    assert w.getCustomWidgetMap() == {}

    w.setItemFactories(include=())

    try:
        # check that an explicit call to setCustomWidgetMap works
        dummy = ("taurus.qt.qtgui.panel.test.test_taurusform._DummyTV", (), {})
        w.setCustomWidgetMap({"DataBase": dummy})
        w.setModel(["tango:sys/database/2", "tango:sys/tg_test/1"])
        qtbot.wait_until(lambda: len(w) == 2, timeout=3200)
        assert type(w[0]) == _DummyTV
        assert type(w[1]) == TaurusValue
        assert w.getCustomWidgetMap() == {"DataBase": dummy}

        # check that the custom widget map can be restored
        w.setCustomWidgetMap({})
        w.setModel(["tango:sys/database/2", "tango:sys/tg_test/1"])
        qtbot.wait_until(lambda: len(w) == 2, timeout=3200)
        assert type(w[0]) == TaurusValue
        assert type(w[1]) == TaurusValue
        assert w.getCustomWidgetMap() == {}
    finally:
        # set model to None as an attempt to avoid problems in atexit()
        w.setModel(None)
        qtbot.wait_until(lambda: len(w) == 0, timeout=3200)


def test_form_itemFactory_loading(qtbot):
    """
    check that the factory loading is robust against unloadable plugins
    and badly-implemented item factories
    """

    w = TaurusForm()
    qtbot.addWidget(w)

    w.setItemFactories(
        include=(_BadEntryPoint, _BadFactory, _DummyItemFactory)
    )
    w.setModel(
        ["eval://localhost/@dummy/'test_itemfactory'",
         "eval://localhost/@dummy/'test_badfactory'",
         "eval:1",
         ]
    )
    qtbot.wait_until(lambda: len(w) == 3, timeout=3200)

    # handled by _DummyItemFactory
    assert type(w[0]) == _DummyTV
    # handled in _BadFactory (even if with wrong return value)
    assert type(w[1]) == _DummyTV
    # errored in _BadFactory, ignored by _DummyItemFactory
    assert type(w[2]) == TaurusValue






#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the set of base class taurus controllers."""

from builtins import str
from builtins import object

import weakref

from taurus.external.qt import Qt

from taurus.core.taurusbasetypes import DataFormat, TaurusEventType

from taurus.qt.qtgui.util import QT_ATTRIBUTE_QUALITY_PALETTE
from taurus.qt.qtgui.util import QT_DEVICE_STATE_PALETTE


__all__ = ["TaurusBaseController", "TaurusAttributeControllerHelper",
           "TaurusScalarAttributeControllerHelper",
           "TaurusConfigurationControllerHelper",
           "updateLabelBackground"]

__docformat__ = 'restructuredtext'


class TaurusBaseController(object):
    """Base class for all taurus controllers"""

    def __init__(self, widget, updateAsPalette=True):
        self._widget = weakref.ref(widget)
        self._updateAsPalette = updateAsPalette
        self._stateObj = None
        self._last_value = None
        self._last_config_value = None
        self._last_error_value = None
        self._setStyle()

    def _setStyle(self):
        pass

    def usePalette(self):
        return self._updateAsPalette

    def widget(self):
        return self._widget()

    def modelObj(self):
        return self.widget().getModelObj()

    attrObj = configObj = deviceObj = modelObj

    def valueObj(self):
        value = self._last_value
        if value is None:
            modelObj = self.modelObj()
            if modelObj is None:
                return None
            value = modelObj.getValueObj()
        return value

    def value(self):
        valueObj = self.valueObj()
        return getattr(valueObj, "value", None)

    def w_value(self):  # TODO: adapt to tep14
        valueObj = self.valueObj()
        return getattr(valueObj, "w_value", None)

    def quality(self):
        valueObj = self.valueObj()
        return getattr(valueObj, "quality", None)

    def state(self):
        return self._stateObj.state

    def getDisplayValue(self, write=False):
        return self.widget().getDisplayValue()

    def handleEvent(self, evt_src, evt_type, evt_value):
        # update the "_last" values only if the event source is the model
        # (it could be the background...)
        if evt_src == self.modelObj():
            if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
                if self._last_value is None:
                    # reset the format so that it gets updated by displayValue
                    self.widget().resetFormat()
                self._last_value = evt_value
            elif evt_type == TaurusEventType.Config:  # TODO: adapt to tep14
                self._last_config_value = evt_value
                self.widget().resetFormat()
            else:
                self._last_error_value = evt_value
                # In case of error, modify the last_value as well
                try:
                    self._last_value = self.modelObj().getValueObj()
                except:
                    self._last_value = None
        self.update()

    def eventReceived(self, evt_src, evt_type, evt_value):
        # should handle the state event here. Because this is invoked by a random
        # thread, we pass it to the widget, which will forward to the proper
        # thread

        #@todo: sometimes we get this method called but self.widget() is None. Check why.
        #       For the moment I just protect it by substituting the following line by the ones after it
        #self.widget().eventReceived(evt_src, evt_type, evt_value)
        w = self.widget()
        if w is not None:
            w.eventReceived(evt_src, evt_type, evt_value)

    def update(self):
        widget = self.widget()
        self._updateConnections(widget)
        self._updateForeground(widget)
        self._updateBackground(widget)
        self._updateToolTip(widget)

    def _needsStateConnection(self):
        return False

    def _updateConnections(self, widget):
        stateObj, newStateObj = self._stateObj, None
        if self._needsStateConnection():
            newStateObj = self.deviceObj()
            if stateObj != newStateObj:
                if stateObj is not None:
                    stateObj.removeListener(self)
                if newStateObj is not None:
                    newStateObj.addListener(self)
        else:
            if stateObj is not None:
                stateObj.removeListener(self)
        self._stateObj = newStateObj

    def _updateForeground(self, widget):
        pass

    def _updateBackground(self, widget):
        pass

    def _updateToolTip(self, widget):
        if widget.getAutoTooltip():
            widget.setToolTip(widget.getFormatedToolTip())


class TaurusAttributeControllerHelper(object):

    def configObj(self):
        return self.attrObj()  # they are the same object since tep14
        # attrObj = self.attrObj()
        # if attrObj is None: return None
        # return attrObj.getConfig()

    def deviceObj(self):
        attrObj = self.attrObj()
        if attrObj is None:
            return None
        return attrObj.getParentObj()


class TaurusScalarAttributeControllerHelper(TaurusAttributeControllerHelper):

    def getDisplayValue(self, write=False):
        valueObj = self.valueObj()
        widget = self.widget()
        if valueObj is None or valueObj.rvalue is None:
            return widget.getDisplayValue()

        format = self.attrObj().data_format
        if format == DataFormat._0D:
            return self._getDisplayValue(widget, valueObj, None, write)

        idx = widget.getModelIndexValue()
        return self._getDisplayValue(widget, valueObj, idx, write)

    def _getDisplayValue(self, widget, valueObj, idx, write):
        try:
            if write:
                value = valueObj.wvalue
            else:
                value = valueObj.rvalue
            if idx is not None and len(idx):
                for i in idx:
                    value = value[i]
            return widget.displayValue(value)
        except Exception as e:
            return widget.getNoneValue()

    def displayValue(self, value):
        if value is None:
            return None
        ret = None
        try:
            if self.isScalar():
                format = self.getFormat()
                if self.isNumeric() and format is not None:
                    format = self.getFormat()
                    ret = self.getFormat() % value
                else:
                    ret = str(value)
            elif self.isSpectrum():
                ret = str(value)
            else:
                ret = str(value)
        except:
            # if cannot calculate value based on the format just return the
            # value
            raise
            ret = str(value)
        return ret


class TaurusConfigurationControllerHelper(object):
    #TODO: Check if this is  used. If so, rename to avoid "Configuration"

    def __init__(self):
        self._configParam = None

    def attrObj(self):
        return self.configObj()  # they are the same object since tep14

    def deviceObj(self):
        attrObj = self.attrObj()
        if attrObj is None:
            return None
        return attrObj.getParentObj()

    @property
    def configParam(self):
        if self._configParam is None:
            self._configParam = self.widget().modelFragmentName or ''
        return self._configParam

    def getDisplayValue(self, write=False):
        widget = self.widget()
        model = self.configObj()
        if model is None:
            return widget.getNoneValue()
        param = self.configParam
        try:
            val = widget.getModelFragmentObj()
            try:
                no_val = getattr(model, "no_" + param)  # TODO: Tango-centric
                if val.lower() == no_val.lower():
                    val = widget.getNoneValue()
            except:
                pass
        except AttributeError:
            if param:
                val = str(param)
                attr = self.attrObj()
                if attr is not None:
                    val = val.replace('<label>', attr.label or '---')
                    val = val.replace('<attr_name>', attr.name or '---')
                    val = val.replace('<attr_fullname>', attr.getFullName() or
                                      '---')
                dev = self.deviceObj()
                if dev is not None:
                    val = val.replace('<dev_fullname>', dev.getFullName() or
                                      '---')
                    val = val.replace('<dev_alias>', dev.getSimpleName() or
                                      '---')
                    val = val.replace('<dev_name>', dev.getNormalName() or
                                      '---')
            else:
                val = widget.getNoneValue()
        except:
            widget.debug("Invalid configuration parameter '%s'" % param)
            val = widget.getNoneValue()
        if val is None:
            val = widget.getNoneValue()
        return val


StyleSheetTemplate = """border-style: outset; border-width: 2px; border-color: {0}; {1}"""


def _updatePaletteColors(widget, bgBrush, fgBrush, frameBrush):
    qt_palette = widget.palette()
    qt_palette.setBrush(Qt.QPalette.Window, bgBrush)
    qt_palette.setBrush(Qt.QPalette.Base, bgBrush)
    qt_palette.setBrush(Qt.QPalette.WindowText, fgBrush)
    qt_palette.setBrush(Qt.QPalette.Light, frameBrush)
    qt_palette.setBrush(Qt.QPalette.Dark, frameBrush)
    widget.setPalette(qt_palette)


def updateLabelBackground(ctrl, widget):
    """Helper method to setup background of taurus labels and lcds"""
    bgRole = widget.bgRole

    if ctrl.usePalette():
        widget.setAutoFillBackground(True)
        if bgRole in ('', 'none', 'None'):
            transparentBrush = Qt.QBrush(Qt.Qt.transparent)
            frameBrush = transparentBrush
            bgBrush, fgBrush = transparentBrush, Qt.QBrush(Qt.Qt.black)
        else:
            frameBrush = Qt.QBrush(Qt.QColor(255, 255, 255, 128))
            bgItem, palette = None, QT_DEVICE_STATE_PALETTE
            if bgRole == 'quality':
                palette = QT_ATTRIBUTE_QUALITY_PALETTE
                bgItem = ctrl.quality()
            elif bgRole == 'state':
                try:
                    bgItem = ctrl.state()
                except AttributeError:
                    pass  # protect against calls with state not instantiated
            elif bgRole == 'value':
                bgItem = ctrl.value()
            else:
                # TODO: this is an *experimental* extension of the bgRole API
                # added in v 4.1.2-alpha. It may change in future versions
                modelObj = widget.getModelObj()
                try:
                    bgItem = modelObj.getFragmentObj(bgRole)
                except:
                    widget.warning('Invalid bgRole "%s"', bgRole)
            bgBrush, fgBrush = palette.qbrush(bgItem)
        _updatePaletteColors(widget, bgBrush, fgBrush, frameBrush)
    else:
        if bgRole in ('', 'none', 'None'):
            ss = StyleSheetTemplate.format("rgba(0,0,0,0)", "")
        else:
            bgItem, palette = None, QT_DEVICE_STATE_PALETTE
            if bgRole == 'quality':
                palette = QT_ATTRIBUTE_QUALITY_PALETTE
                bgItem = ctrl.quality()
            elif bgRole == 'state':
                bgItem = ctrl.state()
            elif bgRole == 'value':
                bgItem = ctrl.value()
            else:
                # TODO: this is an *experimental* extension of the bgRole API
                # added in v 4.1.2-alpha. It may change in future versions
                modelObj = widget.getModelObj()
                try:
                    bgItem = modelObj.getFragmentObj(bgRole)
                except:
                    widget.warning('Invalid bgRole "%s"', bgRole)
            color_ss = palette.qtStyleSheet(bgItem)
            ss = StyleSheetTemplate.format("rgba(255,255,255,128)",  color_ss)
        widget.setStyleSheet(ss)
    widget.update()  # necessary in pyqt <= 4.4



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from taurus.core.util.plugin import selectEntryPoints, EntryPointAlike

import pkg_resources
import random
import string


def _random_string(length, chars=string.ascii_lowercase, prefix='', suffix=''):
    """
    Return a random string with given length, prefix and suffix.
    The random characters are chosen from the chars iterable (lowercase
    letters by default)
    """
    r = ''.join(random.choice(chars) for i in range(length))
    return prefix + r + suffix


def mock_entry_point(lines, group=None, dist_name=None):
    """
    Registers a fake distribution that advertises a group of entry points
    (defined by the `lines argument as a sequence of parseable strings).
    The group name is give by the `group` argument (or generated
    randomly if not given). The fake distribution name is given by the
    `dist_name` argument (or generated randomly if not given).

    The function registers (so they are discoverable, e.g. with
    :method:`pkg_resources.iter_entry_points`) and also returns the mapping
    dictionary.
    """
    if group is None:
        group = _random_string(8, prefix='dummygroup_')
    if dist_name is None:
        dist_name = _random_string(8, prefix='dummydist_')

    # create a dummy pkg_resources distribution based on a clone of taurus
    w = pkg_resources.working_set
    d = w.find(pkg_resources.Requirement.parse('taurus'))
    d = d.clone(project_name=dist_name)
    w.add(d, dist_name)

    # create a fake entry point mapping (with group name testgroup))
    mapping = pkg_resources.EntryPoint.parse_map({group: lines}, dist=d)

    # add mapping to the fake distro so that it can be discovered
    d._ep_map = mapping

    # return the mapping
    return mapping


def test_selectEntryPoints():

    # create a fake entry point mapping (with group name testgroup))
    group = _random_string(8, prefix='dummygroup_')
    names = ['foo1', 'foo2', 'baz1', 'baz2', 'bar1', 'bar2']
    mock_entry_point(
        lines=["{0} = obj_{0}".format(n) for n in names],
        group=group
    )

    # check the selectEntryPoints function
    all = selectEntryPoints(group)
    assert isinstance(all[0], pkg_resources.EntryPoint)
    assert [ep.name for ep in all] == sorted(names)

    none_included = selectEntryPoints(group, include=())
    assert len(none_included) == 0

    all_excluded = selectEntryPoints(group, exclude=('.*',))
    assert len(all_excluded) == 0

    some = selectEntryPoints(
        group,
        exclude=("foo2",),
        include=("bar2", "b.*1", "f.*")
    )
    assert [ep.name for ep in some] == ["bar2", "bar1", "baz1", "foo1"]

    mixed = selectEntryPoints(
        group,
        exclude=("b.*",),
        include=(123, "f.*", EntryPointAlike(321, name="boo"))
    )
    assert [ep.name for ep in mixed] == ["123", "foo1", "foo2", "boo"]
    assert mixed[0].load() == 123
    assert mixed[-1].load() == 321
